# -*- coding: utf-8 -*-
from .. import config
from ..compat import theano, tt, ts, ifelse
from .._constants import *
from .ops import (
    sTOp,
    rTReflectedOp,
    sTReflectedOp,
    dotROp,
    tensordotRzOp,
    FOp,
    spotYlmOp,
    pTOp,
    minimizeOp,
    LDPhysicalOp,
    LimbDarkOp,
    GetClOp,
    RaiseValueErrorOp,
    RaiseValueErrorIfOp,
    OrenNayarOp,
)
from .utils import logger, autocompile, is_tensor, clear_cache
from .math import lazy_math as math
from scipy.special import legendre as LegendreP
import numpy as np
from astropy import units
import os
import exoplanet


# C extensions are not installed on RTD
if os.getenv("READTHEDOCS") == "True":  # pragma: no cover
    _c_ops = None
else:
    from .. import _c_ops


__all__ = ["OpsYlm", "OpsLD", "OpsReflected", "OpsRV", "OpsSystem"]


class OpsYlm(object):
    """Class housing Theano operations for spherical harmonics maps."""

    def __init__(self, ydeg, udeg, fdeg, nw, reflected=False, **kwargs):
        # Ingest kwargs
        self.ydeg = ydeg
        self.udeg = udeg
        self.fdeg = fdeg
        self.deg = ydeg + udeg + fdeg
        self.filter = (fdeg > 0) or (udeg > 0)
        self.nw = nw
        self._reflected = reflected
        self.Ny = (self.ydeg + 1) ** 2

        # Instantiate the C++ Ops
        config.rootHandler.terminator = ""
        logger.info("Pre-computing some matrices... ")
        self._c_ops = _c_ops.Ops(ydeg, udeg, fdeg)
        config.rootHandler.terminator = "\n"
        logger.info("Done.")

        # Solution vectors
        self._sT = sTOp(self._c_ops.sT, self._c_ops.N)
        self._rT = tt.shape_padleft(tt.as_tensor_variable(self._c_ops.rT))
        self._rTA1 = tt.shape_padleft(tt.as_tensor_variable(self._c_ops.rTA1))

        # Change of basis matrices
        self._A = ts.as_sparse_variable(self._c_ops.A)
        self._A1 = ts.as_sparse_variable(self._c_ops.A1)
        self._A1Inv = ts.as_sparse_variable(self._c_ops.A1Inv)

        # Rotation operations
        self._tensordotRz = tensordotRzOp(self._c_ops.tensordotRz)
        self._dotR = dotROp(self._c_ops.dotR)

        # Filter
        # TODO: Make the filter operator sparse
        self._F = FOp(self._c_ops.F, self._c_ops.N, self._c_ops.Ny)

        # Misc
        self._spotYlm = spotYlmOp(
            self._c_ops.spotYlm, self.ydeg, self.nw
        )  # Deprecated
        self._pT = pTOp(self._c_ops.pT, self.deg)
        if self.nw is None:
            if self._reflected:
                self._minimize = minimizeOp(
                    self.unweighted_intensity,
                    self.P,
                    self.ydeg,
                    self.udeg,
                    self.fdeg,
                )
            else:
                self._minimize = minimizeOp(
                    self.intensity, self.P, self.ydeg, self.udeg, self.fdeg
                )
        else:
            # TODO: Implement minimization for spectral maps?
            self._minimize = None
        self._LimbDarkIsPhysical = LDPhysicalOp(_c_ops.nroots)

    @property
    def rT(self):
        return self._rT

    @property
    def rTA1(self):
        return self._rTA1

    @property
    def A(self):
        return self._A

    @property
    def A1(self):
        return self._A1

    @property
    def A1Inv(self):
        return self._A1Inv

    @autocompile
    def sT(self, b, r):
        return self._sT(b, r)

    @autocompile
    def tensordotRz(self, matrix, theta):
        return self._tensordotRz(matrix, theta)

    @autocompile
    def dotR(self, matrix, ux, uy, uz, theta):
        return self._dotR(matrix, ux, uy, uz, theta)

    @autocompile
    def F(self, u, f):
        return self._F(u, f)

    @autocompile
    def spotYlm(self, amp, sigma, lat, lon):
        # Deprecated
        return self._spotYlm(amp, sigma, lat, lon)

    @autocompile
    def pT(self, x, y, z):
        return self._pT(x, y, z)

    @autocompile
    def limbdark_is_physical(self, u):
        """Return True if the limb darkening profile is physical."""
        return self._LimbDarkIsPhysical(u)

    @autocompile
    def get_minimum(self, y):
        """Compute the location and value of the intensity minimum."""
        return self._minimize(y)

    @autocompile
    def X(self, theta, xo, yo, zo, ro, inc, obl, u, f):
        """Compute the light curve design matrix."""
        # Determine shapes
        rows = theta.shape[0]
        cols = self.rTA1.shape[1]
        X = tt.zeros((rows, cols))

        # Compute the occultation mask
        b = tt.sqrt(xo ** 2 + yo ** 2)
        b_rot = tt.ge(b, 1.0 + ro) | tt.le(zo, 0.0) | tt.eq(ro, 0.0)
        b_occ = tt.invert(b_rot)
        i_rot = tt.arange(b.size)[b_rot]
        i_occ = tt.arange(b.size)[b_occ]

        # Compute filter operator
        if self.filter:
            F = self.F(u, f)

        # Rotation operator
        if self.filter:
            rTA1 = ts.dot(tt.dot(self.rT, F), self.A1)
        else:
            rTA1 = self.rTA1
        rTA1 = tt.tile(rTA1, (theta[i_rot].shape[0], 1))
        X = tt.set_subtensor(
            X[i_rot], self.right_project(rTA1, inc, obl, theta[i_rot])
        )

        # Occultation + rotation operator
        sT = self.sT(b[i_occ], ro)
        sTA = ts.dot(sT, self.A)
        theta_z = tt.arctan2(xo[i_occ], yo[i_occ])
        sTAR = self.tensordotRz(sTA, theta_z)
        if self.filter:
            A1InvFA1 = ts.dot(ts.dot(self.A1Inv, F), self.A1)
            sTAR = tt.dot(sTAR, A1InvFA1)
        X = tt.set_subtensor(
            X[i_occ], self.right_project(sTAR, inc, obl, theta[i_occ])
        )

        return X

    @autocompile
    def flux(self, theta, xo, yo, zo, ro, inc, obl, y, u, f):
        """Compute the light curve."""
        return tt.dot(self.X(theta, xo, yo, zo, ro, inc, obl, u, f), y)

    @autocompile
    def P(self, lat, lon):
        """Compute the pixelization matrix, no filters or illumination."""
        # Get the Cartesian points
        xpt, ypt, zpt = self.latlon_to_xyz(lat, lon)

        # Compute the polynomial basis at the point
        pT = self.pT(xpt, ypt, zpt)[:, : (self.ydeg + 1) ** 2]

        # Transform to the Ylm basis
        pTA1 = ts.dot(pT, self.A1)

        # NOTE: The factor of `pi` ensures the correct normalization.
        # This is *different* from the derivation in the paper, but it's
        # due to the fact that the in starry we normalize the spherical
        # harmonics in a slightly strange way (they're normalized so that
        # the integral of Y_{0,0} over the unit sphere is 4, not 4pi).
        # This is useful for thermal light maps, where the flux from a map
        # with Y_{0,0} = 1 is *unity*. But it messes up things for reflected
        # light maps, so we need to account for that here.
        if self._reflected:
            pTA1 *= np.pi

        # We're done
        return pTA1

    @autocompile
    def intensity(self, lat, lon, y, u, f, theta, ld):
        """Compute the intensity at a point or a set of points."""
        # Get the Cartesian points
        xpt, ypt, zpt = self.latlon_to_xyz(lat, lon)

        # Compute the polynomial basis at the point
        pT = self.pT(xpt, ypt, zpt)

        # Transform the map to the polynomial basis
        A1y = ts.dot(self.A1, y)

        # Apply the filter
        if self.filter:
            u0 = tt.zeros_like(u)
            u0 = tt.set_subtensor(u0[0], -1.0)
            A1y = ifelse(
                ld, tt.dot(self.F(u, f), A1y), tt.dot(self.F(u0, f), A1y)
            )

        # Dot the polynomial into the basis
        return tt.dot(pT, A1y)

    @autocompile
    def render(self, res, projection, theta, inc, obl, y, u, f):
        """Render the map on a Cartesian grid."""
        # Compute the Cartesian grid
        xyz = ifelse(
            tt.eq(projection, STARRY_RECTANGULAR_PROJECTION),
            self.compute_rect_grid(res)[-1],
            ifelse(
                tt.eq(projection, STARRY_MOLLWEIDE_PROJECTION),
                self.compute_moll_grid(res)[-1],
                self.compute_ortho_grid(res)[-1],
            ),
        )

        # Compute the polynomial basis
        pT = self.pT(xyz[0], xyz[1], xyz[2])

        # If orthographic, rotate the map to the correct frame
        if self.nw is None:
            Ry = ifelse(
                tt.eq(projection, STARRY_ORTHOGRAPHIC_PROJECTION),
                self.left_project(
                    tt.transpose(tt.tile(y, [theta.shape[0], 1])),
                    inc,
                    obl,
                    theta,
                ),
                tt.transpose(tt.tile(y, [theta.shape[0], 1])),
            )
        else:
            Ry = ifelse(
                tt.eq(projection, STARRY_ORTHOGRAPHIC_PROJECTION),
                self.left_project(y, inc, obl, tt.tile(theta[0], self.nw)),
                y,
            )

        # Change basis to polynomials
        A1Ry = ts.dot(self.A1, Ry)

        # Apply the filter *only if orthographic*
        if self.filter:
            f0 = tt.zeros_like(f)
            f0 = tt.set_subtensor(f0[0], np.pi)
            u0 = tt.zeros_like(u)
            u0 = tt.set_subtensor(u0[0], -1.0)
            A1Ry = ifelse(
                tt.eq(projection, STARRY_ORTHOGRAPHIC_PROJECTION),
                tt.dot(self.F(u, f), A1Ry),
                tt.dot(self.F(u0, f0), A1Ry),
            )

        # Dot the polynomial into the basis
        res = tt.reshape(tt.dot(pT, A1Ry), [res, res, -1])

        # We need the shape to be (nframes, npix, npix)
        return res.dimshuffle(2, 0, 1)

    @autocompile
    def expand_spot(self, amp, sigma, lat, lon):
        """Return the spherical harmonic expansion of a Gaussian spot [DEPRECATED]."""
        return self.spotYlm(amp, sigma, lat, lon)

    @autocompile
    def compute_ortho_grid(self, res):
        """Compute the polynomial basis on the plane of the sky."""
        # NOTE: I think there's a bug in Theano related to
        # tt.mgrid; I get different results depending on whether the
        # function is compiled using `theano.function()` or if it
        # is evaluated using `.eval()`. The small perturbation to `res`
        # is a hacky fix that ensures that `y` and `x` are of the
        # correct length in all cases I've tested.
        dx = 2.0 / (res - 0.01)
        y, x = tt.mgrid[-1:1:dx, -1:1:dx]
        z = tt.sqrt(1 - x ** 2 - y ** 2)
        y = tt.set_subtensor(y[tt.isnan(z)], np.nan)
        x = tt.reshape(x, [1, -1])
        y = tt.reshape(y, [1, -1])
        z = tt.reshape(z, [1, -1])
        lat = tt.reshape(0.5 * np.pi - tt.arccos(y), [1, -1])
        lon = tt.reshape(tt.arctan(x / z), [1, -1])
        return tt.concatenate((lat, lon)), tt.concatenate((x, y, z))

    @autocompile
    def compute_ortho_grid_inc_obl(self, res, inc, obl):
        """Compute the polynomial basis on the plane of the sky, accounting
        for the map inclination and obliquity."""
        # See NOTE on tt.mgrid bug in `compute_ortho_grid`
        dx = 2.0 / (res - 0.01)
        y, x = tt.mgrid[-1:1:dx, -1:1:dx]
        z = tt.sqrt(1 - x ** 2 - y ** 2)
        y = tt.set_subtensor(y[tt.isnan(z)], np.nan)
        x = tt.reshape(x, [1, -1])
        y = tt.reshape(y, [1, -1])
        z = tt.reshape(z, [1, -1])
        Robl = self.RAxisAngle(tt.as_tensor_variable([0.0, 0.0, 1.0]), -obl)
        Rinc = self.RAxisAngle(
            tt.as_tensor_variable([tt.cos(obl), tt.sin(obl), 0.0]),
            -(0.5 * np.pi - inc),
        )
        R = tt.dot(Robl, Rinc)
        xyz = tt.dot(R, tt.concatenate((x, y, z)))
        x = tt.reshape(xyz[0], [1, -1])
        y = tt.reshape(xyz[1], [1, -1])
        z = tt.reshape(xyz[2], [1, -1])
        lat = tt.reshape(0.5 * np.pi - tt.arccos(y), [1, -1])
        lon = tt.reshape(tt.arctan2(x, z), [1, -1])
        return tt.concatenate((lat, lon)), tt.concatenate((x, y, z))

    @autocompile
    def compute_rect_grid(self, res):
        """Compute the polynomial basis on a rectangular lat/lon grid."""
        # See NOTE on tt.mgrid bug in `compute_ortho_grid`
        dx = np.pi / (res - 0.01)
        lat, lon = tt.mgrid[
            -np.pi / 2 : np.pi / 2 : dx, -3 * np.pi / 2 : np.pi / 2 : 2 * dx
        ]
        x = tt.reshape(tt.cos(lat) * tt.cos(lon), [1, -1])
        y = tt.reshape(tt.cos(lat) * tt.sin(lon), [1, -1])
        z = tt.reshape(tt.sin(lat), [1, -1])
        R = self.RAxisAngle(tt.as_tensor_variable([1.0, 0.0, 0.0]), -np.pi / 2)
        return (
            tt.concatenate(
                (
                    tt.reshape(lat, [1, -1]),
                    tt.reshape(lon + 0.5 * np.pi, [1, -1]),
                )
            ),
            tt.dot(R, tt.concatenate((x, y, z))),
        )

    @autocompile
    def compute_moll_grid(self, res):
        """Compute the polynomial basis on a Mollweide grid."""
        # See NOTE on tt.mgrid bug in `compute_ortho_grid`
        dx = 2 * np.sqrt(2) / (res - 0.01)
        y, x = tt.mgrid[
            -np.sqrt(2) : np.sqrt(2) : dx,
            -2 * np.sqrt(2) : 2 * np.sqrt(2) : 2 * dx,
        ]

        # Make points off-grid nan
        a = np.sqrt(2)
        b = 2 * np.sqrt(2)
        y = tt.where((y / a) ** 2 + (x / b) ** 2 <= 1, y, np.nan)

        # https://en.wikipedia.org/wiki/Mollweide_projection
        theta = tt.arcsin(y / np.sqrt(2))
        lat = tt.arcsin((2 * theta + tt.sin(2 * theta)) / np.pi)
        lon0 = 3 * np.pi / 2
        lon = lon0 + np.pi * x / (2 * np.sqrt(2) * tt.cos(theta))

        # Back to Cartesian, this time on the *sky*
        x = tt.reshape(tt.cos(lat) * tt.cos(lon), [1, -1])
        y = tt.reshape(tt.cos(lat) * tt.sin(lon), [1, -1])
        z = tt.reshape(tt.sin(lat), [1, -1])
        R = self.RAxisAngle(tt.as_tensor_variable([1.0, 0.0, 0.0]), -np.pi / 2)
        return (
            tt.concatenate(
                (
                    tt.reshape(lat, (1, -1)),
                    tt.reshape(lon - 1.5 * np.pi, (1, -1)),
                )
            ),
            tt.dot(R, tt.concatenate((x, y, z))),
        )

    @autocompile
    def right_project(self, M, inc, obl, theta):
        r"""Apply the projection operator on the right.

        Specifically, this method returns the dot product :math:`M \cdot R`,
        where ``M`` is an input matrix and ``R`` is the Wigner rotation matrix
        that transforms a spherical harmonic coefficient vector in the
        input frame to a vector in the observer's frame.
        """
        # Trivial case
        if self.ydeg == 0:
            return M

        # Rotate to the sky frame
        # TODO: Do this in a single compound rotation
        M = self.dotR(
            self.dotR(
                self.dotR(
                    M,
                    -tt.cos(obl),
                    -tt.sin(obl),
                    math.to_tensor(0.0),
                    -(0.5 * np.pi - inc),
                ),
                math.to_tensor(0.0),
                math.to_tensor(0.0),
                math.to_tensor(1.0),
                obl,
            ),
            math.to_tensor(1.0),
            math.to_tensor(0.0),
            math.to_tensor(0.0),
            -0.5 * np.pi,
        )

        # Rotate to the correct phase
        if theta.ndim > 0:
            M = self.tensordotRz(M, theta)

        else:
            M = self.dotR(
                M,
                math.to_tensor(0.0),
                math.to_tensor(0.0),
                math.to_tensor(1.0),
                theta,
            )

        # Rotate to the polar frame
        M = self.dotR(
            M,
            math.to_tensor(1.0),
            math.to_tensor(0.0),
            math.to_tensor(0.0),
            0.5 * np.pi,
        )

        return M

    @autocompile
    def left_project(self, M, inc, obl, theta):
        r"""Apply the projection operator on the left.

        Specifically, this method returns the dot product :math:`R \cdot M`,
        where ``M`` is an input matrix and ``R`` is the Wigner rotation matrix
        that transforms a spherical harmonic coefficient vector in the
        input frame to a vector in the observer's frame.

        """
        # Trivial case
        if self.ydeg == 0:
            return M

        # Note that here we are using the fact that R . M = (M^T . R^T)^T
        MT = tt.transpose(M)

        # Rotate to the polar frame
        MT = self.dotR(
            MT,
            math.to_tensor(1.0),
            math.to_tensor(0.0),
            math.to_tensor(0.0),
            -0.5 * np.pi,
        )

        # Rotate to the correct phase
        if theta.ndim > 0:
            MT = self.tensordotRz(MT, -theta)
        else:
            MT = self.dotR(
                MT,
                math.to_tensor(0.0),
                math.to_tensor(0.0),
                math.to_tensor(1.0),
                -theta,
            )

        # Rotate to the sky frame
        # TODO: Do this in a single compound rotation
        MT = self.dotR(
            self.dotR(
                self.dotR(
                    MT,
                    math.to_tensor(1.0),
                    math.to_tensor(0.0),
                    math.to_tensor(0.0),
                    0.5 * np.pi,
                ),
                math.to_tensor(0.0),
                math.to_tensor(0.0),
                math.to_tensor(1.0),
                -obl,
            ),
            -tt.cos(obl),
            -tt.sin(obl),
            math.to_tensor(0.0),
            (0.5 * np.pi - inc),
        )

        return tt.transpose(MT)

    @autocompile
    def set_map_vector(self, vector, inds, vals):
        """Set the elements of the theano map coefficient tensor."""
        res = tt.set_subtensor(vector[inds], vals * tt.ones_like(vector[inds]))
        return res

    @autocompile
    def latlon_to_xyz(self, lat, lon):
        """Convert lat-lon points to Cartesian points."""
        if lat.ndim == 0:
            lat = tt.shape_padleft(lat, 1)
        if lon.ndim == 0:
            lon = tt.shape_padleft(lon, 1)
        R1 = self.RAxisAngle([1.0, 0.0, 0.0], -lat)
        R2 = self.RAxisAngle([0.0, 1.0, 0.0], lon)
        R = tt.batched_dot(R2, R1)
        xyz = tt.transpose(tt.dot(R, [0.0, 0.0, 1.0]))
        return xyz[0], xyz[1], xyz[2]

    @autocompile
    def RAxisAngle(self, axis=[0, 1, 0], theta=0):
        """Wigner axis-angle rotation matrix."""

        def compute(axis=[0, 1, 0], theta=0):
            axis = tt.as_tensor_variable(axis)
            axis /= axis.norm(2)
            cost = tt.cos(theta)
            sint = tt.sin(theta)

            return tt.reshape(
                tt.as_tensor_variable(
                    [
                        cost + axis[0] * axis[0] * (1 - cost),
                        axis[0] * axis[1] * (1 - cost) - axis[2] * sint,
                        axis[0] * axis[2] * (1 - cost) + axis[1] * sint,
                        axis[1] * axis[0] * (1 - cost) + axis[2] * sint,
                        cost + axis[1] * axis[1] * (1 - cost),
                        axis[1] * axis[2] * (1 - cost) - axis[0] * sint,
                        axis[2] * axis[0] * (1 - cost) - axis[1] * sint,
                        axis[2] * axis[1] * (1 - cost) + axis[0] * sint,
                        cost + axis[2] * axis[2] * (1 - cost),
                    ]
                ),
                [3, 3],
            )

        # If theta is a vector, this is a tensor!
        if hasattr(theta, "ndim") and theta.ndim > 0:
            fn = lambda theta, axis: compute(axis=axis, theta=theta)
            R, _ = theano.scan(fn=fn, sequences=[theta], non_sequences=[axis])
            return R
        else:
            return compute(axis=axis, theta=theta)

    def _spot_setup(
        self, spot_pts=1000, spot_eps=1e-9, spot_smoothing=None, spot_fac=300
    ):
        # The default smoothing depends on `ydeg`
        if spot_smoothing is None:
            if self.ydeg < 4:
                spot_smoothing = 0.5
            else:
                spot_smoothing = 2.0 / self.ydeg

        # Check the cache
        try:
            if (
                (spot_pts == self._spot_pts)
                and (spot_eps == self._spot_eps)
                and (spot_smoothing == self._spot_smoothing)
                and (spot_fac == self._spot_fac)
            ):
                return
        except AttributeError:
            pass

        # Update the cached values
        self._spot_pts = spot_pts
        self._spot_eps = spot_eps
        self._spot_smoothing = spot_smoothing
        self._spot_fac = spot_fac

        # Clear the compiled function cache
        clear_cache(self, self.spot)

        # Pre-compute the linalg stuff
        theta = np.linspace(0, np.pi, spot_pts)
        cost = np.cos(theta)
        B = np.hstack(
            [
                np.sqrt(2 * l + 1) * LegendreP(l)(cost).reshape(-1, 1)
                for l in range(self.ydeg + 1)
            ]
        )
        A = np.linalg.solve(B.T @ B + spot_eps * np.eye(self.ydeg + 1), B.T)
        l = np.arange(self.ydeg + 1)
        i = l * (l + 1)
        S = np.exp(-0.5 * i * spot_smoothing ** 2)
        self._spot_Bp = S[:, None] * A
        self._spot_idx = i
        self._spot_theta = theta
        self._spot_fac = spot_fac

    @autocompile
    def spot(self, contrast, radius, lat, lon):

        # Compute unit-intensity spot at (0, 0)
        z = self._spot_fac * (self._spot_theta - radius)
        b = 1.0 / (1.0 + tt.exp(-z)) - 1.0
        yT = tt.zeros((1, self.Ny))
        yT = tt.set_subtensor(yT[:, self._spot_idx], tt.dot(self._spot_Bp, b))

        # Rotate in latitude then in longitude
        yT = self.dotR(yT, np.array(1.0), np.array(0.0), np.array(0.0), lat)
        yT = self.dotR(yT, np.array(0.0), np.array(1.0), np.array(0.0), -lon)

        # Reshape and we're done
        if self.nw is None:
            y = yT.reshape((-1,)) * contrast
        else:
            y = yT.reshape((-1, 1)) * tt.reshape(contrast, (1, -1))

        return y


class OpsLD(object):
    """Class housing Theano operations for limb-darkened maps."""

    def __init__(self, ydeg, udeg, fdeg, nw, reflected=False, **kwargs):
        # Sanity checks
        assert ydeg == fdeg == 0
        assert reflected is False

        # Ingest kwargs
        self.udeg = udeg
        self.nw = nw

        # Set up the ops
        self._get_cl = GetClOp()
        self._limbdark = LimbDarkOp()
        self._LimbDarkIsPhysical = LDPhysicalOp(_c_ops.nroots)

    @autocompile
    def limbdark_is_physical(self, u):
        """Return True if the limb darkening profile is physical."""
        return self._LimbDarkIsPhysical(u)

    @autocompile
    def intensity(self, mu, u):
        """Compute the intensity at a set of points."""
        if self.udeg == 0:
            mu_f = tt.reshape(mu, (-1,))
            intensity = tt.ones_like(mu_f)
            intensity = tt.set_subtensor(intensity[tt.isnan(mu_f)], np.nan)
            return intensity
        else:
            basis = tt.reshape(1.0 - mu, (-1, 1)) ** np.arange(self.udeg + 1)
            return -tt.dot(basis, u)

    @autocompile
    def flux(self, xo, yo, zo, ro, u):
        """Compute the light curve."""
        # Initialize flat light curve
        flux = tt.ones_like(xo)

        # Compute the occultation mask
        b = tt.sqrt(xo ** 2 + yo ** 2)
        b_occ = tt.invert(tt.ge(b, 1.0 + ro) | tt.le(zo, 0.0) | tt.eq(ro, 0.0))
        i_occ = tt.arange(b.size)[b_occ]

        # Get the Agol `c` coefficients
        c = self._get_cl(u)
        if self.udeg == 0:
            c_norm = c / (np.pi * c[0])
        else:
            c_norm = c / (np.pi * (c[0] + 2 * c[1] / 3))

        # Compute the occultation flux
        los = zo[i_occ]
        r = ro * tt.ones_like(los)
        flux = tt.set_subtensor(
            flux[i_occ], self._limbdark(c_norm, b[i_occ], r, los)[0]
        )
        return flux

    @autocompile
    def X(self, theta, xo, yo, zo, ro, inc, obl, u, f):
        """
        Convenience function for integration of limb-darkened maps
        with the ``System`` class. The design matrix for limb-darkened
        maps is just a column vector equal to the total flux, since the
        spherical harmonic coefficient vector is ``[1.0]``.

        """
        flux = self.flux(xo, yo, zo, ro, u)
        X = tt.reshape(flux, (-1, 1))
        return X

    @autocompile
    def render(self, res, projection, theta, inc, obl, y, u, f):
        """Render the map on a Cartesian grid."""
        nframes = tt.shape(theta)[0]
        image = self.render_ld(res, u)
        return tt.tile(image, (nframes, 1, 1))

    @autocompile
    def render_ld(self, res, u):
        """Simplified version of `render` w/o the extra params.

        The method `render` requires a bunch of dummy params for
        compatibility with the `System` class. This method is a
        convenience method for use in the `Map` class.
        """
        # See NOTE on tt.mgrid bug in `compute_ortho_grid`
        dx = 2.0 / (res - 0.01)
        y, x = tt.mgrid[-1:1:dx, -1:1:dx]

        x = tt.reshape(x, [1, -1])
        y = tt.reshape(y, [1, -1])
        mu = tt.sqrt(1 - x ** 2 - y ** 2)

        # Compute the intensity
        intensity = self.intensity(mu, u)

        # We need the shape to be (nframes, npix, npix)
        return tt.reshape(intensity, (1, res, res))

    @autocompile
    def set_map_vector(self, vector, inds, vals):
        """Set the elements of the theano map coefficient tensor."""
        res = tt.set_subtensor(vector[inds], vals * tt.ones_like(vector[inds]))
        return res


class OpsRV(OpsYlm):
    """Class housing Theano operations for radial velocity maps."""

    @autocompile
    def compute_rv_filter(self, inc, obl, veq, alpha):
        """Compute the radial velocity field Ylm multiplicative filter."""
        # Define some angular quantities
        cosi = tt.cos(inc)
        sini = tt.sin(inc)
        cosl = tt.cos(obl)
        sinl = tt.sin(obl)
        A = sini * cosl
        B = sini * sinl
        C = cosi

        # Compute the Ylm expansion of the RV field
        return (
            tt.reshape(
                [
                    0,
                    veq
                    * np.sqrt(3)
                    * B
                    * (-(A ** 2) * alpha - B ** 2 * alpha - C ** 2 * alpha + 5)
                    / 15,
                    0,
                    veq
                    * np.sqrt(3)
                    * A
                    * (-(A ** 2) * alpha - B ** 2 * alpha - C ** 2 * alpha + 5)
                    / 15,
                    0,
                    0,
                    0,
                    0,
                    0,
                    veq * alpha * np.sqrt(70) * B * (3 * A ** 2 - B ** 2) / 70,
                    veq
                    * alpha
                    * 2
                    * np.sqrt(105)
                    * C
                    * (-(A ** 2) + B ** 2)
                    / 105,
                    veq
                    * alpha
                    * np.sqrt(42)
                    * B
                    * (A ** 2 + B ** 2 - 4 * C ** 2)
                    / 210,
                    0,
                    veq
                    * alpha
                    * np.sqrt(42)
                    * A
                    * (A ** 2 + B ** 2 - 4 * C ** 2)
                    / 210,
                    veq * alpha * 4 * np.sqrt(105) * A * B * C / 105,
                    veq * alpha * np.sqrt(70) * A * (A ** 2 - 3 * B ** 2) / 70,
                ],
                [-1],
            )
            * np.pi
        )

    @autocompile
    def rv(self, theta, xo, yo, zo, ro, inc, obl, y, u, veq, alpha):
        """Compute the observed radial velocity anomaly."""
        # Compute the velocity-weighted intensity
        f = self.compute_rv_filter(inc, obl, veq, alpha)
        Iv = self.flux(theta, xo, yo, zo, ro, inc, obl, y, u, f)

        # Compute the inverse of the intensity
        f0 = tt.zeros_like(f)
        f0 = tt.set_subtensor(f0[0], np.pi)
        I = self.flux(theta, xo, yo, zo, ro, inc, obl, y, u, f0)
        invI = tt.ones((1,)) / I
        invI = tt.where(tt.isinf(invI), 0.0, invI)

        # The RV signal is just the product
        return Iv * invI


class OpsReflected(OpsYlm):
    """Class housing Theano operations for reflected light maps."""

    def __init__(self, *args, **kwargs):
        super(OpsReflected, self).__init__(*args, reflected=True, **kwargs)
        self._rT = rTReflectedOp(self._c_ops.rTReflected, self._c_ops.N)
        self._sT = sTReflectedOp(self._c_ops.sTReflected, self._c_ops.N)
        self._A1Big = ts.as_sparse_variable(self._c_ops.A1Big)

        # Compute grid on unit disk with ~source_npts points
        source_npts = kwargs.get("source_npts", 1)
        if source_npts <= 1:
            self.source_dx = np.array([0.0])
            self.source_dy = np.array([0.0])
            self.source_dz = np.array([0.0])
            self.source_npts = 1
        else:
            N = int(2 + np.sqrt(source_npts * 4 / np.pi))
            dx = np.linspace(-1, 1, N)
            dx, dy = np.meshgrid(dx, dx)
            dz = 1 - dx ** 2 - dy ** 2
            self.source_dx = dx[dz > 0].flatten()
            self.source_dy = dy[dz > 0].flatten()
            self.source_dz = dz[dz > 0].flatten()
            self.source_npts = len(self.source_dx)

        # NOTE: dz is *negative*, since the source is actually
        # *closer* to the body than in the point approximation
        self.source_dx = tt.as_tensor_variable(self.source_dx)
        self.source_dy = tt.as_tensor_variable(self.source_dy)
        self.source_dz = tt.as_tensor_variable(-self.source_dz)

        # Oren-Nayar (1994) intensity profile (for rendering)
        self._OrenNayar = OrenNayarOp(self._c_ops.OrenNayarPolynomial)
        self._pTON94 = pTOp(self._c_ops.pT, _c_ops.STARRY_OREN_NAYAR_DEG)

    @property
    def A1Big(self):
        return self._A1Big

    @autocompile
    def rT(self, b, sigr):
        return self._rT(b, sigr)[0]

    @autocompile
    def sT(self, b, theta, bo, ro, sigr):
        return self._sT(b, theta, bo, ro, sigr)[0]

    @autocompile
    def intensity(
        self,
        lat,
        lon,
        y,
        u,
        f,
        xs,
        ys,
        zs,
        Rs,
        theta,
        ld,
        sigr,
        on94_exact,
        illuminate,
    ):
        """Compute the intensity at a series of lat-lon points on the surface."""
        # Get the Cartesian points
        xpt, ypt, zpt = self.latlon_to_xyz(lat, lon)

        # Compute the polynomial basis at the point
        pT = self.pT(xpt, ypt, zpt)

        # Transform the map to the polynomial basis
        A1y = ts.dot(self.A1, y)

        # Apply the filter
        if self.filter:
            u0 = tt.zeros_like(u)
            u0 = tt.set_subtensor(u0[0], -1.0)
            A1y = ifelse(
                ld, tt.dot(self.F(u, f), A1y), tt.dot(self.F(u0, f), A1y)
            )

        # Dot the polynomial into the basis
        intensity = tt.shape_padright(tt.dot(pT, A1y))

        # Weight the intensity by the illumination
        xyz = tt.concatenate(
            (
                tt.reshape(xpt, [1, -1]),
                tt.reshape(ypt, [1, -1]),
                tt.reshape(zpt, [1, -1]),
            )
        )
        I = self.compute_illumination(xyz, xs, ys, zs, Rs, sigr, on94_exact)

        # Add an extra dimension for the wavelength
        if self.nw is not None:
            I = tt.shape_padaxis(I, 1)

        # Weight the image by the illumination
        # NOTE: The factor of `pi` ensures the correct normalization.
        # This is *different* from the derivation in the paper, but it's
        # due to the fact that the in starry we normalize the spherical
        # harmonics in a slightly strange way (they're normalized so that
        # the integral of Y_{0,0} over the unit sphere is 4, not 4pi).
        # This is useful for thermal light maps, where the flux from a map
        # with Y_{0,0} = 1 is *unity*. But it messes up things for reflected
        # light maps, so we need to account for that here.
        intensity = tt.switch(
            tt.isnan(intensity),
            intensity,
            ifelse(
                illuminate, intensity * I, np.pi * intensity * tt.ones_like(I)
            ),
        )

        return intensity

    @autocompile
    def unweighted_intensity(self, lat, lon, y, u, f, theta, ld):
        """
        Compute the intensity in the absence of an illumination source
        (i.e., the albedo).

        """
        # Get the Cartesian points
        xpt, ypt, zpt = self.latlon_to_xyz(lat, lon)

        # Compute the polynomial basis at the point
        pT = self.pT(xpt, ypt, zpt)

        # Transform the map to the polynomial basis
        A1y = ts.dot(self.A1, y)

        # Apply the filter
        if self.filter:
            u0 = tt.zeros_like(u)
            u0 = tt.set_subtensor(u0[0], -1.0)
            A1y = ifelse(
                ld, tt.dot(self.F(u, f), A1y), tt.dot(self.F(u0, f), A1y)
            )

        # Dot the polynomial into the basis.
        # NOTE: The factor of `pi` ensures the correct normalization.
        # This is *different* from the derivation in the paper, but it's
        # due to the fact that the in starry we normalize the spherical
        # harmonics in a slightly strange way (they're normalized so that
        # the integral of Y_{0,0} over the unit sphere is 4, not 4pi).
        # This is useful for thermal light maps, where the flux from a map
        # with Y_{0,0} = 1 is *unity*. But it messes up things for reflected
        # light maps, so we need to account for that here.
        return np.pi * tt.dot(pT, A1y)

    @autocompile
    def X_point_source(
        self, theta, xs, ys, zs, xo, yo, zo, ro, inc, obl, u, f, sigr
    ):
        """Compute the light curve design matrix for a point source."""
        # Determine shapes
        rows = theta.shape[0]
        cols = (self.ydeg + 1) ** 2
        X = tt.zeros((rows, cols))

        # Compute the occultation mask
        bo = tt.sqrt(xo ** 2 + yo ** 2)
        b_rot = tt.ge(bo, 1.0 + ro) | tt.le(zo, 0.0) | tt.eq(ro, 0.0)
        b_occ = tt.invert(b_rot)
        i_rot = tt.arange(bo.size)[b_rot]
        i_occ = tt.arange(bo.size)[b_occ]

        # Compute filter operator
        if self.filter:
            F = self.F(u, f)

        # Terminator
        r2 = xs ** 2 + ys ** 2 + zs ** 2
        b_term = -zs / tt.sqrt(r2)
        theta_term = tt.arctan2(xo, yo) - tt.arctan2(xs, ys)

        # Rotation operator
        rT = self.rT(b_term[i_rot], sigr)
        if self.filter:
            rTA1 = ts.dot(tt.dot(rT, F), self.A1)
        else:
            rTA1 = ts.dot(rT, self.A1)
        theta_z = tt.arctan2(xs[i_rot], ys[i_rot])
        rTA1Rz = self.tensordotRz(rTA1, theta_z)
        X = tt.set_subtensor(
            X[i_rot], self.right_project(rTA1Rz, inc, obl, theta[i_rot])
        )

        # Occultation + rotation operator
        sT = self.sT(b_term[i_occ], theta_term[i_occ], bo[i_occ], ro, sigr)
        sTA = ts.dot(sT, self.A)
        theta_z = tt.arctan2(xo[i_occ], yo[i_occ])
        sTAR = self.tensordotRz(sTA, theta_z)
        if self.filter:
            A1InvFA1 = ts.dot(ts.dot(self.A1Inv, F), self.A1)
            sTAR = tt.dot(sTAR, A1InvFA1)
        X = tt.set_subtensor(
            X[i_occ], self.right_project(sTAR, inc, obl, theta[i_occ])
        )

        # Weight by the distance to the source.
        # NOTE: In the paper, we divide by `pi` at this step. But the way
        # we coded things up, starry maps are already implicitly pi-normalized
        # (via the change-of-basis matrix A1), so we don't need to do that here.
        X /= tt.shape_padright(r2)

        # We're done
        return X

    @autocompile
    def X(self, theta, xs, ys, zs, Rs, xo, yo, zo, ro, inc, obl, u, f, sigr):
        """Compute the light curve design matrix."""

        if self.source_npts == 1:

            return self.X_point_source(
                theta, xs, ys, zs, xo, yo, zo, ro, inc, obl, u, f, sigr
            )

        else:

            # The effective size of the star as seen by the planet
            # is smaller. Only include points
            # that fall on this smaller disk.
            rs = tt.sqrt(xs ** 2 + ys ** 2 + zs ** 2)
            Reff = Rs * tt.sqrt(1 - ((Rs - 1) / rs) ** 2)
            dx = tt.shape_padright(Reff) * self.source_dx
            dy = tt.shape_padright(Reff) * self.source_dy
            # Note that the star is *closer* to the planet, hence the - sign
            dz = -tt.sqrt(Rs ** 2 - dx ** 2 - dy ** 2)

            # Compute the illumination for each point on the source disk
            X = self.X_point_source(
                tt.reshape(
                    tt.shape_padright(theta) + tt.zeros_like(dx), (-1,)
                ),
                tt.reshape(tt.shape_padright(xs) + dx, (-1,)),
                tt.reshape(tt.shape_padright(ys) + dy, (-1,)),
                tt.reshape(tt.shape_padright(zs) + dz, (-1,)),
                tt.reshape(tt.shape_padright(xo) + tt.zeros_like(dx), (-1,)),
                tt.reshape(tt.shape_padright(yo) + tt.zeros_like(dx), (-1,)),
                tt.reshape(tt.shape_padright(zo) + tt.zeros_like(dx), (-1,)),
                ro,
                inc,
                obl,
                u,
                f,
                sigr,
            )
            X = tt.reshape(X, (X.shape[0], -1))

            # Point source approximation
            X0 = self.X_point_source(
                theta, xs, ys, zs, xo, yo, zo, ro, inc, obl, u, f, sigr
            )
            X0 = tt.reshape(X0, (X0.shape[0], -1))

            # Average over each profile if Rs != 0
            return ifelse(
                Rs > 0,
                ifelse(
                    tt.shape(theta)[0] > 0,
                    (
                        tt.sum(
                            tt.reshape(
                                X, (tt.shape(theta)[0], self.source_npts, -1)
                            ),
                            axis=1,
                        )
                        / self.source_npts
                    ),
                    tt.zeros_like(X),
                ),
                X0,
            )

    @autocompile
    def flux(
        self, theta, xs, ys, zs, Rs, xo, yo, zo, ro, inc, obl, y, u, f, sigr
    ):
        """Compute the reflected light curve."""
        return tt.dot(
            self.X(
                theta, xs, ys, zs, Rs, xo, yo, zo, ro, inc, obl, u, f, sigr
            ),
            y,
        )

    @autocompile
    def flux_point_source(
        self, theta, xs, ys, zs, xo, yo, zo, ro, inc, obl, y, u, f, sigr
    ):
        """Compute the reflected light curve for a point source."""
        return tt.dot(
            self.X_point_source(
                theta, xs, ys, zs, xo, yo, zo, ro, inc, obl, u, f, sigr
            ),
            y,
        )

    @autocompile
    def render(
        self,
        res,
        projection,
        illuminate,
        theta,
        inc,
        obl,
        y,
        u,
        f,
        xs,
        ys,
        zs,
        Rs,
        sigr,
        on94_exact,
    ):
        """Render the map on a Cartesian grid."""
        # Compute the Cartesian grid
        xyz = ifelse(
            tt.eq(projection, STARRY_RECTANGULAR_PROJECTION),
            self.compute_rect_grid(res)[-1],
            ifelse(
                tt.eq(projection, STARRY_MOLLWEIDE_PROJECTION),
                self.compute_moll_grid(res)[-1],
                self.compute_ortho_grid(res)[-1],
            ),
        )

        # If orthographic, rotate the map to the correct frame
        if self.nw is None:
            Ry = ifelse(
                tt.eq(projection, STARRY_ORTHOGRAPHIC_PROJECTION),
                self.left_project(
                    tt.transpose(tt.tile(y, [theta.shape[0], 1])),
                    inc,
                    obl,
                    theta,
                ),
                tt.transpose(tt.tile(y, [theta.shape[0], 1])),
            )
        else:
            Ry = ifelse(
                tt.eq(projection, STARRY_ORTHOGRAPHIC_PROJECTION),
                self.left_project(y, inc, obl, tt.tile(theta[0], self.nw)),
                y,
            )

        # Transform to polynomials
        A1Ry = ts.dot(self.A1, Ry)

        # Apply the filter *only if orthographic*
        f0 = tt.zeros_like(f)
        f0 = tt.set_subtensor(f0[0], np.pi)
        u0 = tt.zeros_like(u)
        u0 = tt.set_subtensor(u0[0], -1.0)
        A1Ry = ifelse(
            tt.eq(projection, STARRY_ORTHOGRAPHIC_PROJECTION),
            tt.dot(self.F(u, f), A1Ry),
            tt.dot(self.F(u0, f0), A1Ry),
        )

        # Dot the polynomial into the basis
        pT = self.pT(xyz[0], xyz[1], xyz[2])
        image = tt.dot(pT, A1Ry)

        # Compute the illumination profile
        I = self.compute_illumination(xyz, xs, ys, zs, Rs, sigr, on94_exact)

        # Add an extra dimension for the wavelength
        if self.nw is not None:
            I = tt.repeat(I, self.nw, 1)

        # Weight the image by the illumination
        # NOTE: The factor of `pi` ensures the correct normalization.
        # This is *different* from the derivation in the paper, but it's
        # due to the fact that the in starry we normalize the spherical
        # harmonics in a slightly strange way (they're normalized so that
        # the integral of Y_{0,0} over the unit sphere is 4, not 4pi).
        # This is useful for thermal light maps, where the flux from a map
        # with Y_{0,0} = 1 is *unity*. But it messes up things for reflected
        # light maps, so we need to account for that here.
        image = ifelse(
            illuminate,
            tt.switch(tt.isnan(image), image, image * I),
            np.pi * image,
        )

        # We need the shape to be (nframes, npix, npix)
        return tt.reshape(image, [res, res, -1]).dimshuffle(2, 0, 1)

    @autocompile
    def compute_illumination_point_source(
        self, xyz, xs, ys, zs, sigr, on94_exact
    ):
        """Compute the illumination profile for a point source."""
        # Get cos(theta_i)
        x = tt.shape_padright(xyz[0])
        y = tt.shape_padright(xyz[1])
        z = tt.shape_padright(xyz[2])
        r2 = xs ** 2 + ys ** 2 + zs ** 2
        b = -zs / tt.sqrt(r2)  # semi-minor axis of terminator
        invsr = 1.0 / tt.sqrt(xs ** 2 + ys ** 2)
        cosw = ys * invsr
        sinw = -xs * invsr
        xrot = x * cosw + y * sinw
        yrot = -x * sinw + y * cosw
        bc = tt.sqrt(1.0 - b ** 2)
        cos_thetai = bc * yrot - b * z

        # Check for special cases
        cos_thetai = tt.switch(
            tt.eq(tt.abs_(b), 1.0),
            tt.switch(
                tt.eq(b, 1.0), tt.zeros_like(cos_thetai), z  # midnight  # noon
            ),
            cos_thetai,
        )
        # Set night to zero
        cos_thetai = tt.switch(
            tt.gt(cos_thetai, 0.0), cos_thetai, tt.zeros_like(cos_thetai)
        )

        # Lambertian intensity
        I_lamb = cos_thetai

        # Polynomial approximation to the Oren-Nayar (1994)
        # intensity with the nightside masked
        pT = self._pTON94(xyz[0], xyz[1], xyz[2])
        theta = -tt.arctan2(xs, ys)
        p = self._OrenNayar(b, theta, sigr)
        I_on94_approx = tt.dot(pT, p)
        I_on94_approx = tt.switch(
            tt.gt(cos_thetai, 0.0), I_on94_approx, tt.zeros_like(I_on94_approx)
        )

        # "Exact" Oren-Nayar (1994) intensity from their Equation (30)
        f1 = -b / z - cos_thetai
        f2 = -b / cos_thetai - z
        f = cos_thetai * tt.maximum(0, tt.minimum(f1, f2))
        sig2 = sigr ** 2
        A = 1.0 - 0.5 * sig2 / (sig2 + 0.33)
        B = 0.45 * sig2 / (sig2 + 0.09)
        I_on94_exact = A * cos_thetai + B * f

        # Select the function we want
        I = ifelse(
            sigr > 0, ifelse(on94_exact, I_on94_exact, I_on94_approx), I_lamb
        )

        # Weight by the distance to the source.
        # NOTE: In the paper, we divide by `pi` at this step. But the way
        # we coded things up, starry maps are already implicitly pi-normalized
        # (via the change-of-basis matrix A1), so we don't need to do that here.
        I /= tt.shape_padleft(r2)
        return I

    @autocompile
    def compute_illumination(self, xyz, xs, ys, zs, Rs, sigr, on94_exact):
        """Compute the illumination profile when rendering maps."""

        if self.source_npts == 1:

            return self.compute_illumination_point_source(
                xyz, xs, ys, zs, sigr, on94_exact
            )

        else:

            # The effective size of the star as seen by the planet
            # is smaller. Only include points
            # that fall on this smaller disk.
            rs = tt.sqrt(xs ** 2 + ys ** 2 + zs ** 2)
            Reff = Rs * tt.sqrt(1 - ((Rs - 1) / rs) ** 2)
            dx = tt.shape_padright(Reff) * self.source_dx
            dy = tt.shape_padright(Reff) * self.source_dy
            # Note that the star is *closer* to the planet, hence the - sign
            dz = -tt.sqrt(Rs ** 2 - dx ** 2 - dy ** 2)

            # Compute the illumination for each point on the source disk
            I = self.compute_illumination_point_source(
                xyz,
                tt.reshape(tt.shape_padright(xs) + dx, (-1,)),
                tt.reshape(tt.shape_padright(ys) + dy, (-1,)),
                tt.reshape(tt.shape_padright(zs) + dz, (-1,)),
                sigr,
                on94_exact,
            )
            I = tt.reshape(I, (-1, tt.shape(xs)[0], self.source_npts))

            # Average over each profile
            return tt.sum(I, axis=2) / self.source_npts


class OpsSystem(object):
    """Class housing ops for modeling Keplerian systems."""

    def __init__(
        self,
        primary,
        secondaries,
        reflected=False,
        rv=False,
        light_delay=False,
        texp=None,
        oversample=7,
        order=0,
    ):
        # System members
        self.primary = primary
        self.secondaries = secondaries
        self._reflected = reflected
        self._rv = rv
        self.nw = self.primary._map.nw
        self.light_delay = light_delay
        self.texp = texp
        self.oversample = oversample
        self.order = order

    @autocompile
    def position(
        self,
        t,
        pri_m,
        pri_t0,
        sec_m,
        sec_t0,
        sec_porb,
        sec_ecc,
        sec_w,
        sec_Omega,
        sec_iorb,
    ):
        """Compute the Cartesian positions of all bodies."""
        orbit = exoplanet.orbits.KeplerianOrbit(
            period=sec_porb,
            t0=sec_t0,
            incl=sec_iorb,
            ecc=sec_ecc,
            omega=sec_w,
            Omega=sec_Omega,
            m_planet=sec_m,
            m_star=pri_m,
            r_star=1.0,  # doesn't matter
        )
        # Position of the primary
        if len(self.secondaries) > 1:
            x_pri, y_pri, z_pri = tt.sum(orbit.get_star_position(t), axis=-1)
        else:
            x_pri, y_pri, z_pri = orbit.get_star_position(t)

        # Positions of the secondaries
        try:
            x_sec, y_sec, z_sec = orbit.get_planet_position(
                t, light_delay=self.light_delay
            )
        except TypeError:
            if self.light_delay:
                logger.warn(
                    "This version of `exoplanet` does not model light delays."
                )
            x_sec, y_sec, z_sec = orbit.get_planet_position(t)

        # Concatenate them
        x = tt.transpose(tt.concatenate((x_pri, x_sec), axis=-1))
        y = tt.transpose(tt.concatenate((y_pri, y_sec), axis=-1))
        z = tt.transpose(tt.concatenate((z_pri, z_sec), axis=-1))

        return x, y, z

    @autocompile
    def X(
        self,
        t,
        pri_r,
        pri_m,
        pri_prot,
        pri_t0,
        pri_theta0,
        pri_amp,
        pri_inc,
        pri_obl,
        pri_u,
        pri_f,
        sec_r,
        sec_m,
        sec_prot,
        sec_t0,
        sec_theta0,
        sec_porb,
        sec_ecc,
        sec_w,
        sec_Omega,
        sec_iorb,
        sec_amp,
        sec_inc,
        sec_obl,
        sec_u,
        sec_f,
        sec_sigr,
    ):
        """Compute the system light curve design matrix."""
        # Exposure time integration?
        if self.texp != 0.0:

            texp = tt.as_tensor_variable(self.texp)
            oversample = int(self.oversample)
            oversample += 1 - oversample % 2
            stencil = np.ones(oversample)

            # Construct the exposure time integration stencil
            if self.order == 0:
                dt = np.linspace(-0.5, 0.5, 2 * oversample + 1)[1:-1:2]
            elif self.order == 1:
                dt = np.linspace(-0.5, 0.5, oversample)
                stencil[1:-1] = 2
            elif self.order == 2:
                dt = np.linspace(-0.5, 0.5, oversample)
                stencil[1:-1:2] = 4
                stencil[2:-1:2] = 2
            else:
                raise ValueError("Parameter `order` must be <= 2")
            stencil /= np.sum(stencil)

            if texp.ndim == 0:
                dt = texp * dt
            else:
                dt = tt.shape_padright(texp) * dt
            t = tt.shape_padright(t) + dt
            t = tt.reshape(t, (-1,))

        # Compute the relative positions of all bodies
        orbit = exoplanet.orbits.KeplerianOrbit(
            period=sec_porb,
            t0=sec_t0,
            incl=sec_iorb,
            ecc=sec_ecc,
            omega=sec_w,
            Omega=sec_Omega,
            m_planet=sec_m,
            m_star=pri_m,
            r_star=pri_r,
        )
        try:
            x, y, z = orbit.get_relative_position(
                t, light_delay=self.light_delay
            )
        except TypeError:
            if self.light_delay:
                logger.warn(
                    "This version of `exoplanet` does not model light delays."
                )
            x, y, z = orbit.get_relative_position(t)

        # Get all rotational phases
        pri_prot = ifelse(
            tt.eq(pri_prot, 0.0), math.to_tensor(np.inf), pri_prot
        )
        theta_pri = (2 * np.pi) / pri_prot * (t - pri_t0) + pri_theta0
        sec_prot = tt.switch(
            tt.eq(sec_prot, 0.0), math.to_tensor(np.inf), sec_prot
        )
        theta_sec = (2 * np.pi) / tt.shape_padright(sec_prot) * (
            tt.shape_padleft(t) - tt.shape_padright(sec_t0)
        ) + tt.shape_padright(sec_theta0)

        # Compute all the phase curves
        phase_pri = pri_amp * self.primary.map.ops.X(
            theta_pri,
            tt.zeros_like(t),
            tt.zeros_like(t),
            tt.zeros_like(t),
            math.to_tensor(0.0),
            pri_inc,
            pri_obl,
            pri_u,
            pri_f,
        )
        if self._reflected:
            phase_sec = [
                pri_amp
                * sec_amp[i]
                * sec.map.ops.X(
                    theta_sec[i],
                    -x[:, i] / sec_r[i],
                    -y[:, i] / sec_r[i],
                    -z[:, i] / sec_r[i],
                    pri_r / sec_r[i],  # scaled source radius
                    tt.zeros_like(x[:, i]),
                    tt.zeros_like(x[:, i]),
                    tt.zeros_like(x[:, i]),
                    math.to_tensor(0.0),  # occultor of zero radius
                    sec_inc[i],
                    sec_obl[i],
                    sec_u[i],
                    sec_f[i],
                    sec_sigr[i],
                )
                for i, sec in enumerate(self.secondaries)
            ]
        else:
            phase_sec = [
                sec_amp[i]
                * sec.map.ops.X(
                    theta_sec[i],
                    -x[:, i],
                    -y[:, i],
                    -z[:, i],
                    math.to_tensor(0.0),  # occultor of zero radius
                    sec_inc[i],
                    sec_obl[i],
                    sec_u[i],
                    sec_f[i],
                )
                for i, sec in enumerate(self.secondaries)
            ]

        # Compute any occultations
        occ_pri = tt.zeros_like(phase_pri)
        occ_sec = [tt.zeros_like(ps) for ps in phase_sec]

        # Compute the period if we were given a semi-major axis
        sec_porb = tt.switch(
            tt.eq(sec_porb, 0.0),
            (G_grav * (pri_m + sec_m) * sec_porb ** 2 / (4 * np.pi ** 2))
            ** (1.0 / 3),
            sec_porb,
        )

        # Compute transits across the primary
        for i, _ in enumerate(self.secondaries):
            xo = x[:, i] / pri_r
            yo = y[:, i] / pri_r
            zo = z[:, i] / pri_r
            ro = sec_r[i] / pri_r
            b = tt.sqrt(xo ** 2 + yo ** 2)
            b_occ = tt.invert(
                tt.ge(b, 1.0 + ro) | tt.le(zo, 0.0) | tt.eq(ro, 0.0)
            )
            idx = tt.arange(b.shape[0])[b_occ]
            occ_pri = tt.set_subtensor(
                occ_pri[idx],
                occ_pri[idx]
                + pri_amp
                * self.primary.map.ops.X(
                    theta_pri[idx],
                    xo[idx],
                    yo[idx],
                    zo[idx],
                    ro,
                    pri_inc,
                    pri_obl,
                    pri_u,
                    pri_f,
                )
                - phase_pri[idx],
            )

        # Compute occultations by the primary
        for i, sec in enumerate(self.secondaries):

            xo = -x[:, i] / sec_r[i]
            yo = -y[:, i] / sec_r[i]
            zo = -z[:, i] / sec_r[i]
            ro = pri_r / sec_r[i]
            b = tt.sqrt(xo ** 2 + yo ** 2)
            b_occ = tt.invert(
                tt.ge(b, 1.0 + ro) | tt.le(zo, 0.0) | tt.eq(ro, 0.0)
            )
            idx = tt.arange(b.shape[0])[b_occ]
            if self._reflected:
                occ_sec[i] = tt.set_subtensor(
                    occ_sec[i][idx],
                    occ_sec[i][idx]
                    + pri_amp
                    * sec_amp[i]
                    * sec.map.ops.X(
                        theta_sec[i, idx],
                        xo[idx],  # the primary is both the source...
                        yo[idx],
                        zo[idx],
                        ro,
                        xo[idx],  # ... and the occultor
                        yo[idx],
                        zo[idx],
                        ro,
                        sec_inc[i],
                        sec_obl[i],
                        sec_u[i],
                        sec_f[i],
                        sec_sigr[i],
                    )
                    - phase_sec[i][idx],
                )
            else:
                occ_sec[i] = tt.set_subtensor(
                    occ_sec[i][idx],
                    occ_sec[i][idx]
                    + sec_amp[i]
                    * sec.map.ops.X(
                        theta_sec[i, idx],
                        xo[idx],
                        yo[idx],
                        zo[idx],
                        ro,
                        sec_inc[i],
                        sec_obl[i],
                        sec_u[i],
                        sec_f[i],
                    )
                    - phase_sec[i][idx],
                )

        # Compute secondary-secondary occultations
        for i, sec in enumerate(self.secondaries):
            for j, _ in enumerate(self.secondaries):
                if i == j:
                    continue
                xo = (-x[:, i] + x[:, j]) / sec_r[i]
                yo = (-y[:, i] + y[:, j]) / sec_r[i]
                zo = (-z[:, i] + z[:, j]) / sec_r[i]
                ro = sec_r[j] / sec_r[i]
                b = tt.sqrt(xo ** 2 + yo ** 2)
                b_occ = tt.invert(
                    tt.ge(b, 1.0 + ro) | tt.le(zo, 0.0) | tt.eq(ro, 0.0)
                )
                idx = tt.arange(b.shape[0])[b_occ]
                if self._reflected:
                    xs = -x[:, i] / sec_r[i]
                    ys = -y[:, i] / sec_r[i]
                    zs = -z[:, i] / sec_r[i]
                    occ_sec[i] = tt.set_subtensor(
                        occ_sec[i][idx],
                        occ_sec[i][idx]
                        + sec_amp[i]
                        * pri_amp
                        * sec.map.ops.X(
                            theta_sec[i, idx],
                            xs[idx],  # the primary is the source
                            ys[idx],
                            zs[idx],
                            pri_r / sec_r[i],
                            xo[idx],  # another secondary is the occultor
                            yo[idx],
                            zo[idx],
                            ro,
                            sec_inc[i],
                            sec_obl[i],
                            sec_u[i],
                            sec_f[i],
                            sec_sigr[i],
                        )
                        - phase_sec[i][idx],
                    )
                else:
                    occ_sec[i] = tt.set_subtensor(
                        occ_sec[i][idx],
                        occ_sec[i][idx]
                        + sec_amp[i]
                        * sec.map.ops.X(
                            theta_sec[i, idx],
                            xo[idx],
                            yo[idx],
                            zo[idx],
                            ro,
                            sec_inc[i],
                            sec_obl[i],
                            sec_u[i],
                            sec_f[i],
                        )
                        - phase_sec[i][idx],
                    )

        # Concatenate the design matrices
        X_pri = phase_pri + occ_pri
        X_sec = [ps + os for ps, os in zip(phase_sec, occ_sec)]
        X = tt.horizontal_stack(X_pri, *X_sec)

        # Sum and return
        if self.texp == 0.0:

            return X

        else:

            stencil = tt.shape_padright(tt.shape_padleft(stencil, 1), 1)
            return tt.sum(
                stencil * tt.reshape(X, (-1, self.oversample, X.shape[1])),
                axis=1,
            )

    @autocompile
    def rv(
        self,
        t,
        pri_r,
        pri_m,
        pri_prot,
        pri_t0,
        pri_theta0,
        pri_amp,
        pri_inc,
        pri_obl,
        pri_y,
        pri_u,
        pri_alpha,
        pri_veq,
        sec_r,
        sec_m,
        sec_prot,
        sec_t0,
        sec_theta0,
        sec_porb,
        sec_ecc,
        sec_w,
        sec_Omega,
        sec_iorb,
        sec_amp,
        sec_inc,
        sec_obl,
        sec_y,
        sec_u,
        sec_alpha,
        sec_sigr,
        sec_veq,
        keplerian,
    ):
        """Compute the observed system radial velocity (RV maps only)."""
        # TODO: This method is currently very inefficient, as it
        # calls `X` twice per call and instantiates an `orbit`
        # instance up to three separate times per call. We should
        # re-code the logic from `X()` in here to optimize it.

        # Compute the RV filter
        pri_f = self.primary.map.ops.compute_rv_filter(
            pri_inc, pri_obl, pri_veq, pri_alpha
        )
        sec_f = tt.as_tensor_variable(
            [
                sec.map.ops.compute_rv_filter(
                    sec_inc[k], sec_obl[k], sec_veq[k], sec_alpha[k]
                )
                for k, sec in enumerate(self.secondaries)
            ]
        )

        # Compute the identity filter
        pri_f0 = tt.zeros_like(pri_f)
        pri_f0 = tt.set_subtensor(pri_f0[0], np.pi)
        sec_f0 = tt.as_tensor_variable([pri_f0 for sec in self.secondaries])

        # Compute the two design matrices
        X = self.X(
            t,
            pri_r,
            pri_m,
            pri_prot,
            pri_t0,
            pri_theta0,
            pri_amp,
            pri_inc,
            pri_obl,
            pri_u,
            pri_f,
            sec_r,
            sec_m,
            sec_prot,
            sec_t0,
            sec_theta0,
            sec_porb,
            sec_ecc,
            sec_w,
            sec_Omega,
            sec_iorb,
            sec_amp,
            sec_inc,
            sec_obl,
            sec_u,
            sec_f,
            sec_sigr,
        )

        X0 = self.X(
            t,
            pri_r,
            pri_m,
            pri_prot,
            pri_t0,
            pri_theta0,
            pri_amp,
            pri_inc,
            pri_obl,
            pri_u,
            pri_f0,
            sec_r,
            sec_m,
            sec_prot,
            sec_t0,
            sec_theta0,
            sec_porb,
            sec_ecc,
            sec_w,
            sec_Omega,
            sec_iorb,
            sec_amp,
            sec_inc,
            sec_obl,
            sec_u,
            sec_f0,
            sec_sigr,
        )

        # Get the indices of X corresponding to each body
        pri_inds = np.arange(0, self.primary.map.Ny, dtype=int)
        sec_inds = [None for sec in self.secondaries]
        n = self.primary.map.Ny
        for i, sec in enumerate(self.secondaries):
            sec_inds[i] = np.arange(n, n + sec.map.Ny, dtype=int)
            n += sec.map.Ny

        # Compute the integral of the velocity-weighted intensity
        Iv = tt.as_tensor_variable(
            [tt.dot(X[:, pri_inds], pri_y)]
            + [
                tt.dot(X[:, sec_inds[n]], sec_y[n])
                for n in range(len(self.secondaries))
            ]
        )

        # Compute the inverse of the integral of the intensity
        invI = tt.as_tensor_variable(
            [tt.ones((1,)) / tt.dot(X0[:, pri_inds], pri_y)]
            + [
                tt.ones((1,)) / tt.dot(X0[:, sec_inds[n]], sec_y[n])
                for n in range(len(self.secondaries))
            ]
        )
        invI = tt.where(tt.isinf(invI), 0.0, invI)

        # The RV anomaly is just the product
        rv = Iv * invI

        # Compute the Keplerian RV
        orbit = exoplanet.orbits.KeplerianOrbit(
            period=sec_porb,
            t0=sec_t0,
            incl=sec_iorb,
            ecc=sec_ecc,
            omega=sec_w,
            Omega=sec_Omega,
            m_planet=sec_m,
            m_star=pri_m,
            r_star=pri_r,
        )
        return ifelse(
            keplerian,
            tt.inc_subtensor(
                rv[1:],
                tt.transpose(
                    orbit.get_radial_velocity(
                        t, output_units=units.m / units.s
                    )
                ),
            ),
            rv,
        )

    @autocompile
    def render(
        self,
        t,
        res,
        pri_r,
        pri_m,
        pri_prot,
        pri_t0,
        pri_theta0,
        pri_inc,
        pri_obl,
        pri_y,
        pri_u,
        pri_f,
        sec_r,
        sec_m,
        sec_prot,
        sec_t0,
        sec_theta0,
        sec_porb,
        sec_ecc,
        sec_w,
        sec_Omega,
        sec_iorb,
        sec_inc,
        sec_obl,
        sec_y,
        sec_u,
        sec_f,
        sec_sigr,
    ):
        """Render all of the bodies in the system."""
        # Compute the relative positions of all bodies
        orbit = exoplanet.orbits.KeplerianOrbit(
            period=sec_porb,
            t0=sec_t0,
            incl=sec_iorb,
            ecc=sec_ecc,
            omega=sec_w,
            Omega=sec_Omega,
            m_planet=sec_m,
            m_star=pri_m,
            r_star=pri_r,
        )
        try:
            x, y, z = orbit.get_relative_position(
                t, light_delay=self.light_delay
            )
        except TypeError:
            if self.light_delay:
                logger.warn(
                    "This version of `exoplanet` does not model light delays."
                )
            x, y, z = orbit.get_relative_position(t)

        # Get all rotational phases
        pri_prot = ifelse(
            tt.eq(pri_prot, 0.0), math.to_tensor(np.inf), pri_prot
        )
        theta_pri = (2 * np.pi) / pri_prot * (t - pri_t0) + pri_theta0
        sec_prot = tt.switch(
            tt.eq(sec_prot, 0.0), math.to_tensor(np.inf), sec_prot
        )
        theta_sec = (2 * np.pi) / tt.shape_padright(sec_prot) * (
            tt.shape_padleft(t) - tt.shape_padright(sec_t0)
        ) + tt.shape_padright(sec_theta0)

        # Compute all the maps
        img_pri = self.primary.map.ops.render(
            res,
            STARRY_ORTHOGRAPHIC_PROJECTION,
            theta_pri,
            pri_inc,
            pri_obl,
            pri_y,
            pri_u,
            pri_f,
        )
        if self._reflected:
            img_sec = tt.as_tensor_variable(
                [
                    sec.map.ops.render(
                        res,
                        STARRY_ORTHOGRAPHIC_PROJECTION,
                        1,
                        theta_sec[i],
                        sec_inc[i],
                        sec_obl[i],
                        sec_y[i],
                        sec_u[i],
                        sec_f[i],
                        -x[:, i],
                        -y[:, i],
                        -z[:, i],
                        pri_r / sec_r[i],
                        sec_sigr[i],
                        0,  # use approx Oren-Nayar intensity
                    )
                    for i, sec in enumerate(self.secondaries)
                ]
            )
        else:
            img_sec = tt.as_tensor_variable(
                [
                    sec.map.ops.render(
                        res,
                        STARRY_ORTHOGRAPHIC_PROJECTION,
                        theta_sec[i],
                        sec_inc[i],
                        sec_obl[i],
                        sec_y[i],
                        sec_u[i],
                        sec_f[i],
                    )
                    for i, sec in enumerate(self.secondaries)
                ]
            )

        # Return the images and secondary orbital positions
        return img_pri, img_sec, x, y, z



# -*- coding: utf-8 -*-
"""
System linear solve tests.

"""
import starry
import numpy as np
from scipy.linalg import cho_solve
from scipy.stats import multivariate_normal
import pytest
import itertools

# Parameter combinations we'll test
vals = ["scalar", "vector", "matrix", "cholesky"]
woodbury = [False, True]
solve_inputs = itertools.product(vals, vals)
lnlike_inputs = itertools.product(vals, vals, woodbury)


@pytest.fixture
def model():
    class Model:
        def __init__(self):
            # Instantiate a star with a dipole map
            A = starry.Primary(starry.Map(ydeg=1), prot=0.0)
            amp_true = 0.75
            y_true = np.array([1, 0.1, 0.2, 0.3])
            inc_true = 60
            A.map.amp = amp_true
            A.map[1, :] = y_true[1:]
            A.map.inc = inc_true

            # Instantiate two transiting planets with different longitudes of
            # ascending node. This ensures there's no null space!
            b = starry.Secondary(
                starry.Map(amp=0), porb=1.0, r=0.1, t0=-0.05, Omega=30.0
            )
            c = starry.Secondary(
                starry.Map(amp=0), porb=1.0, r=0.1, t0=0.05, Omega=-30.0
            )
            sys = starry.System(A, b, c)

            # Generate a synthetic light curve with just a little noise
            t = np.linspace(-0.1, 0.1, 100)
            flux = sys.flux(t)
            sigma = 1e-5
            np.random.seed(1)
            flux += np.random.randn(len(t)) * sigma

            # Store
            self.A = A
            self.b = b
            self.c = c
            self.sys = sys
            self.t = t
            self.flux = flux
            self.sigma = sigma
            self.amp_true = amp_true
            self.y_true = y_true
            self.inc_true = inc_true

    return Model()


@pytest.mark.parametrize("L,C", solve_inputs)
def test_solve(L, C, model):
    # Place a generous prior on the map coefficients
    if L == "scalar":
        model.A.map.set_prior(L=1)
    elif L == "vector":
        model.A.map.set_prior(L=np.ones(model.A.map.Ny))
    elif L == "matrix":
        model.A.map.set_prior(L=np.eye(model.A.map.Ny))
    elif L == "cholesky":
        model.A.map.set_prior(cho_L=np.eye(model.A.map.Ny))

    # Provide the dataset
    if C == "scalar":
        model.sys.set_data(model.flux, C=model.sigma ** 2)
    elif C == "vector":
        model.sys.set_data(
            model.flux, C=np.ones(len(model.flux)) * model.sigma ** 2
        )
    elif C == "matrix":
        model.sys.set_data(
            model.flux, C=np.eye(len(model.flux)) * model.sigma ** 2
        )
    elif C == "cholesky":
        model.sys.set_data(
            model.flux, cho_C=np.eye(len(model.flux)) * model.sigma
        )

    # Solve the linear problem
    model.A.map.inc = model.inc_true
    mu, cho_cov = model.sys.solve(t=model.t)

    # Ensure the likelihood of the true value is close to that of
    # the MAP solution
    cov = cho_cov.dot(cho_cov.T)
    LnL0 = multivariate_normal.logpdf(mu, mean=mu, cov=cov)
    LnL = multivariate_normal.logpdf(
        model.amp_true * model.y_true, mean=mu, cov=cov
    )
    assert LnL0 - LnL < 5.00

    # Check that we can draw from the posterior
    model.sys.draw()


@pytest.mark.parametrize("L,C,woodbury", lnlike_inputs)
def test_lnlike(L, C, woodbury, model):
    # Place a generous prior on the map coefficients
    if L == "scalar":
        model.A.map.set_prior(L=1)
    elif L == "vector":
        model.A.map.set_prior(L=np.ones(model.A.map.Ny))
    elif L == "matrix":
        model.A.map.set_prior(L=np.eye(model.A.map.Ny))
    elif L == "cholesky":
        model.A.map.set_prior(cho_L=np.eye(model.A.map.Ny))

    # Provide the dataset
    if C == "scalar":
        model.sys.set_data(model.flux, C=model.sigma ** 2)
    elif C == "vector":
        model.sys.set_data(
            model.flux, C=np.ones(len(model.flux)) * model.sigma ** 2
        )
    elif C == "matrix":
        model.sys.set_data(
            model.flux, C=np.eye(len(model.flux)) * model.sigma ** 2
        )
    elif C == "cholesky":
        model.sys.set_data(
            model.flux, cho_C=np.eye(len(model.flux)) * model.sigma
        )

    # Compute the marginal log likelihood for different secondari radii
    rs = [0.05, 0.075, 0.1, 0.125, 0.15]
    ll = np.zeros_like(rs)
    for i, r in enumerate(rs):
        model.b.r = r
        ll[i] = model.sys.lnlike(t=model.t, woodbury=woodbury)

    # Verify that we get the correct radius
    assert rs[np.argmax(ll)] == 0.1
    assert np.allclose(ll[np.argmax(ll)], 981.9091)  # benchmarked


def test_solve_with_zero_degree_primary():
    """
    Test for https://github.com/rodluger/starry/issues/253.

    """
    # Random data
    t = np.linspace(-0.1, 0.1, 100)
    flux = np.random.randn(len(t))
    sigma = 1.0

    # Instantiate a primary with ydeg = 0
    star_0 = starry.Primary(starry.Map(ydeg=0))
    star_0.map.set_prior(L=1)

    # Instantiate a secondary with ydeg > 0
    star_1 = starry.Secondary(starry.Map(ydeg=1), porb=1.0)

    # The system
    sys = starry.System(star_0, star_1)

    # Place a generous prior on the map coefficients
    star_1.map.set_prior(L=1)
    sys.set_data(flux, C=sigma ** 2)

    # Solve the linear problem
    mu, cho_cov = sys.solve(t=t)

    # Check that it didn't fail horribly
    assert not np.any(np.isnan(mu))
    assert not np.any(np.isnan(np.tril(cho_cov)))



import numpy as np
import os

np.seterr(invalid="ignore")

__all__ = ["generate_header"]


def get_f_exact(x, y, z, b):
    r"""
    Return the expression

    .. math::

        f \equiv
        \mathrm{max}\left(0, \cos(\phi_r - \phi_i)\right)
        \sin\alpha \tan\beta

    from Equation (30) in Oren & Nayar (1994) as a function
    of the Cartesian coordinates on the sky-projected sphere
    seen at a phase where the semi-minor axis of the terminator
    is `b`.

    """
    bc = np.sqrt(1 - b ** 2)
    ci = bc * y - b * z
    f1 = -b / z - ci
    f2 = -b / ci - z
    f = np.maximum(0, np.minimum(f1, f2))
    return f


def get_ijk(n):
    """Get the exponents of x, y, z i the nth term of the polynomial basis."""
    l = int(np.floor(np.sqrt(n)))
    m = n - l * l - l
    mu = l - m
    nu = l + m
    if mu % 2 == 0:
        i = mu // 2
        j = nu // 2
        k = 0
    else:
        i = (mu - 1) // 2
        j = (nu - 1) // 2
        k = 1
    return i, j, k


def poly_basis(x, y, z, deg):
    """Return the polynomial basis evaluated at `x`, `y`, `z`."""
    N = (deg + 1) ** 2
    B = np.zeros((len(x * y * z), N))
    for n in range(N):
        i, j, k = get_ijk(n)
        B[:, n] = x ** i * y ** j * z ** k
    return B


def design_matrix(x, y, z, b, deg, Nb):
    """
    Return the x-y-z-b-bc Vandermonde design matrix.
    
    NOTE: The lowest power of `b` is *ONE*, since
    we need `f = 0` eveywhere when `b = 0` for
    a smooth transition to Lambertian at crescent
    phase.
    """
    N = (deg + 1) ** 2
    u = 0
    X = np.zeros((len(y * z * b), N * Nb ** 2))
    bc = np.sqrt(1 - b ** 2)
    B = poly_basis(x, y, z, deg)
    for n in range(N):
        for p in range(1, Nb + 1):
            for q in range(Nb):
                X[:, u] = B[:, n] * b ** p * bc ** q
                u += 1
    return X


def index_of(i, j, k, p, q, deg, Nb):
    """
    Return the index in `w` corresponding to a certain term.
    
    Not at all optimized! In fact, very much the opposite.

    """
    idx = 0
    for n in range((deg + 1) ** 2):
        i0, j0, k0 = get_ijk(n)
        for p0 in range(1, Nb + 1):
            for q0 in range(Nb):
                if (
                    (i0 == i)
                    and (j0 == j)
                    and (k0 == k)
                    and (p0 == p)
                    and (q0 == q)
                ):
                    return idx
                idx += 1
    raise IndexError("Invalid polynomial index!")


def get_w(deg=5, Nb=4, res=100, prior_var=1e2):
    """
    Return the coefficients of the 5D fit to `f`
    in `x`, `y`, `z`, `b`, and `bc`.

    We fit the function `f` (see above) with a 
    polynomial of order `deg0 = deg - 1` in `x`, `y`, and `z`
    and `Nb0 = Nb - 1` in `b` and `bc`. We then multiply the
    result by `cos(theta_i)`, which is just a polynomial of
    order 1 in `y`, `z`, `b`, and `bc`.

    The final polynomial has order `deg` and `Nb` in the
    Cartesian and terminator coordinates, respectively, and is a fit
    to the function

    .. math::

        \cos\theta_i
        \mathrm{max}\left(0, \cos(\phi_r - \phi_i)\right)
        \sin\alpha \tan\beta

    from Equation (30) in Oren & Nayar (1994).

    """
    # Degrees before multiplying by cos(theta_i)
    deg0 = deg - 1
    Nb0 = Nb - 1

    # Construct a 3D grid in (x, y, b)
    bgrid = np.linspace(-1, 0, res)
    xygrid = np.linspace(-1, 1, res)
    x, y, b = np.meshgrid(xygrid, xygrid, bgrid)
    z = np.sqrt(1 - x ** 2 - y ** 2)
    idx = np.isfinite(z) & (y > b * np.sqrt(1 - x ** 2))
    x = x[idx]
    y = y[idx]
    z = z[idx]
    b = b[idx]

    # Compute the exact `f` function on this grid
    f = get_f_exact(x, y, z, b)

    # Construct the design matrix for fitting
    X = design_matrix(x, y, z, b, deg=deg0, Nb=Nb0)

    # "Data" inverse covariance. Make the errorbars large
    # when cos(theta_i) is small, since the intensity there
    # is small anyways.
    cinv = np.ones_like(f)
    bc = np.sqrt(1 - b ** 2)
    ci = bc * y - b * z
    cinv *= ci ** 2

    # Prior inverse covariance. Make odd powers of x have
    # *very* narrow priors centered on zero, since the
    # function should be symmetric about the y axis.
    # We'll explicitly zero out these coefficients below.
    PInv = np.eye((deg0 + 1) ** 2 * Nb0 ** 2)
    u = 0
    for n in range((deg0 + 1) ** 2):
        i, _, _ = get_ijk(n)
        if (i % 2) != 0:
            inv_var = 1e15
        else:
            inv_var = 1 / prior_var
        for p in range(1, Nb0 + 1):
            for q in range(Nb0):
                PInv[u, u] = inv_var
                u += 1

    # Solve the L2 problem
    XTCInv = X.T * cinv
    w0 = np.linalg.solve(XTCInv.dot(X) + PInv, XTCInv.dot(f),)

    # Zero out really tiny values.
    w0[np.abs(w0) < 1e-10] = 0.0

    # Now multiply by cos_thetai = bc * y - b * z.
    # The powers of x, y, z, b, bc are
    # i, j, k, p, q, respectively
    w = np.zeros((deg + 1) ** 2 * Nb ** 2)
    for n in range((deg0 + 1) ** 2):
        i, j, k = get_ijk(n)
        for p in range(1, Nb0 + 1):
            for q in range(Nb0):
                idx = index_of(i, j, k, p, q, deg0, Nb0)
                w[index_of(i, j + 1, k, p, q + 1, deg, Nb)] += w0[idx]

                if k == 0:
                    w[index_of(i, j, k + 1, p + 1, q, deg, Nb)] -= w0[idx]
                else:
                    # transform z^2 --> 1 - x^2 - y^2
                    w[index_of(i, j, 0, p + 1, q, deg, Nb)] -= w0[idx]
                    w[index_of(i + 2, j, 0, p + 1, q, deg, Nb)] += w0[idx]
                    w[index_of(i, j + 2, 0, p + 1, q, deg, Nb)] += w0[idx]
    return w


def generate_header(deg=5, Nb=4, res=100, nperline=3):
    print("Generating `oren_nayar.h`...")
    assert deg >= 1, "deg must be >= 1"
    assert Nb >= 1, "Nb must be >= 1"
    w = get_w(deg=deg, Nb=Nb, res=res)
    N = len(w) - (len(w) % nperline)
    string = "static const double STARRY_OREN_NAYAR_COEFFS[] = {\n"
    lines = w[:N].reshape(-1, nperline)
    last_line = w[N:]
    for i, line in enumerate(lines):
        string += ", ".join(["{:22.15e}".format(value) for value in line])
        if (i < len(lines) - 1) or (len(last_line)):
            string += ","
        string += "\n"
    string += ", ".join(["{:22.15e}".format(value) for value in last_line])
    string += "};"
    with open(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "oren_nayar.h"
        ),
        "w",
    ) as f:
        print("#ifndef _STARRY_REFLECTED_OREN_NAYAR_H_", file=f)
        print("#define _STARRY_REFLECTED_OREN_NAYAR_H_", file=f)
        print("", file=f)
        print("#define STARRY_OREN_NAYAR_DEG {}".format(deg), file=f)
        print("#define STARRY_OREN_NAYAR_N {}".format((deg + 1) ** 2), file=f)
        print("#define STARRY_OREN_NAYAR_NB {}".format(Nb), file=f)
        print("", file=f)
        print(string, file=f)
        print("#endif", file=f)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Check the quality of the fit for various values of `b`.

    # Get the polynomial coefficients
    deg = 5
    Nb = 4
    w = get_w(deg=deg, Nb=Nb)

    # Grid the surface
    res = 300
    xygrid = np.linspace(-1, 1, res)
    x, y = np.meshgrid(xygrid, xygrid)
    x = x.reshape(-1)
    y = y.reshape(-1)
    z = np.sqrt(1 - x ** 2 - y ** 2)

    # Compare for several values of b
    nimg = 6
    fig, ax = plt.subplots(nimg, 4, figsize=(6, 8))
    for axis in ax.flatten():
        axis.axis("off")
    for i, b in enumerate(np.linspace(-1, 0, nimg, endpoint=False)):

        # Illumination profile
        bc = np.sqrt(1 - b ** 2)
        ci = bc * y - b * z

        # Compute the exact `f` function on this grid
        f = ci * get_f_exact(x, y, z, b)

        # Get our approximation
        X = design_matrix(x, y, z, b, deg=deg, Nb=Nb)
        fapprox = X.dot(w)

        # Mask the nightside & reshape
        idx = np.isfinite(z) & (y < b * np.sqrt(1 - x ** 2))
        f[idx] = 0
        fapprox[idx] = 0
        f = f.reshape(res, res)
        fapprox = fapprox.reshape(res, res)

        # Plot
        vmin = 0
        vmax = 1
        ax[i, 0].imshow(
            f, origin="lower", extent=(-1, 1, -1, 1), vmin=vmin, vmax=vmax
        )
        ax[i, 1].imshow(
            fapprox,
            origin="lower",
            extent=(-1, 1, -1, 1),
            vmin=vmin,
            vmax=vmax,
        )
        ax[i, 2].imshow(
            f - fapprox,
            origin="lower",
            extent=(-1, 1, -1, 1),
            vmin=-0.1,
            vmax=0.1,
            cmap="RdBu",
        )

        bins = np.linspace(-0.1, 0.1, 50)
        ax[i, 3].hist((f - fapprox).flatten(), bins=bins)
        ax[i, 3].set_xlim(-0.1, 0.1)

    plt.show()



# -*- coding: utf-8 -*-
"""
Functions to test limb darkened occultations.

"""
import starry
import numpy as np


def test_quadratic():
    """Test quadratic limb darkening."""
    # Occultation params
    xo = np.linspace(-1.5, 1.5, 1000)
    yo = 0.3
    ro = 0.1

    # LD params
    u1 = 0.5
    u2 = 0.25

    # Limb-darkened map
    map1 = starry.Map(udeg=2)
    map1[1] = u1
    map1[2] = u2
    flux1 = map1.flux(xo=xo, yo=yo, ro=ro)

    # Emulate it with a Ylm map (Equation 38 in Luger et al. 2019)
    map2 = starry.Map(ydeg=2)
    y00 = 2.0 * np.sqrt(np.pi) / 3.0 * (3.0 - 3.0 * u1 - 4.0 * u2)
    y10 = 2.0 * np.sqrt(np.pi / 3.0) * (u1 + 2.0 * u2)
    y20 = -4.0 / 3.0 * np.sqrt(np.pi / 5.0) * u2
    y = np.array([y00, 0, y10, 0, 0, 0, y20, 0, 0]) / (
        np.pi * (1.0 - u1 / 3.0 - u2 / 6.0)
    )
    flux2 = map2.design_matrix(xo=xo, yo=yo, ro=ro).dot(y)
    flux2 *= np.sqrt(np.pi) / 2  # add in the starry normalization

    assert np.allclose(flux1, flux2)



# -*- coding: utf-8 -*-
"""
Test multi-wavelength maps.

"""
import starry
import pytest
import itertools

# Params combinations
ydeg = [2]
udeg = [0]
nw = [5]
rv = [False, True]
reflected = [False, True]
params = itertools.product(ydeg, udeg, nw, rv, reflected)

# Reflected light + radial velocity is not implemented
params = [p for p in params if not (p[3] and p[4])]


@pytest.fixture(scope="class", params=params)
def map(request):
    ydeg, udeg, nw, rv, reflected = request.param
    map = starry.Map(ydeg=ydeg, udeg=udeg, nw=nw, reflected=reflected, rv=rv)
    map.reflected = reflected
    return map


class TestShapes:
    """Test the flux and intensity return value shapes."""

    def test_flux(self, map):
        assert map.flux().eval().shape == (1, map.nw)

    def test_flux_vector(self, map):
        assert map.flux(xo=[0, 1, 2]).eval().shape == (3, map.nw)

    def test_intensity(self, map):
        if map.reflected:
            assert map.intensity().eval().shape == (1, map.nw, 1)
        else:
            assert map.intensity().eval().shape == (1, map.nw)

    def test_intensity_vector(self, map):
        if map.reflected:
            assert map.intensity(lat=[0, 30, 60]).eval().shape == (
                3,
                map.nw,
                1,
            )
        else:
            assert map.intensity(lat=[0, 30, 60]).eval().shape == (3, map.nw)

    def test_intensity_matrix(self, map):
        if map.reflected:
            assert map.intensity(lat=[0, 30, 60], xs=[1, 2]).eval().shape == (
                3,
                map.nw,
                2,
            )
        else:
            pass

    def test_rv(self, map):
        if hasattr(map, "rv"):
            assert map.rv().eval().shape == (1, map.nw)
        else:
            pass

    def test_rv_vector(self, map):
        if hasattr(map, "rv"):
            assert map.rv(xo=[0, 1, 2]).eval().shape == (3, map.nw)
        else:
            pass

    def test_render(self, map):
        res = 100
        map.render(res=res).eval().shape == (map.nw, res, res)



# -*- coding: utf-8 -*-
"""Test map rotation.

"""
import starry
import numpy as np


def test_rotate():
    map = starry.Map(1)
    map[1, 1] = 1
    map.rotate([0, 0, 1], 90.0)
    assert np.allclose(map.y.eval(), [1, 1, 0, 0])

    map = starry.Map(1)
    map[1, -1] = 1
    map.rotate([0, 0, 1], 90.0)
    assert np.allclose(map.y.eval(), [1, 0, 0, -1])


def test_rotate_spectral():
    map = starry.Map(1, nw=2)
    map[1, 1, 0] = 1
    map[1, -1, 1] = 1
    map.rotate([0, 0, 1], 90.0)
    assert np.allclose(map.y.eval(), [[1, 1], [1, 0], [0, 0], [0, -1]])



# -*- coding: utf-8 -*-
"""
There's a bug in tt.mgrid that causes different
behavior whether it's compiled or not. We implemented
some hacks in `starry` to circumvent this.
See docstring of `compute_ortho_grid` in "core.py"

"""
import starry
import pytest
from starry.compat import tt
import itertools
import numpy as np


map = starry.Map(1)


@pytest.mark.parametrize("compile", [True, False])
def test_ortho_grid(compile):
    for res in np.arange(30, 101):
        if compile:
            (lat, lon), (x, y, z) = map.ops.compute_ortho_grid(res)
        else:
            latlon, xyz = map.ops.compute_ortho_grid(
                tt.as_tensor_variable(res)
            )
            x, y, z = xyz.eval()
        assert len(x) == res ** 2



# -*- coding: utf-8 -*-
"""
Linear solve / likelihood tests.

"""
import starry
import numpy as np
from scipy.linalg import cho_solve
from scipy.stats import multivariate_normal
import pytest
import itertools

# Parameter combinations we'll test
vals = ["scalar", "vector", "matrix", "cholesky"]
woodbury = [False, True]
solve_inputs = itertools.product(vals, vals)
lnlike_inputs = itertools.product(vals, vals, woodbury)

# Instantiate a dipole map
map = starry.Map(ydeg=1, reflected=True)
amp_true = 0.75
inc_true = 60
y_true = np.array([1, 0.1, 0.2, 0.3])
map.amp = amp_true
map[1, :] = y_true[1:]
map.inc = inc_true

# Generate a synthetic light curve with just a little noise
theta = np.linspace(0, 360, 100)
phi = 3.5 * theta
xs = np.cos(phi * np.pi / 180)
ys = 0.1 * np.cos(phi * np.pi / 180)
zs = np.sin(phi * np.pi / 180)
kwargs = dict(theta=theta, xs=xs, ys=ys, zs=zs)
flux = map.flux(**kwargs).eval()
sigma = 1e-5
np.random.seed(1)
flux += np.random.randn(len(theta)) * sigma


@pytest.mark.parametrize("L,C", solve_inputs)
def test_solve(L, C):
    # Place a generous prior on the map coefficients
    if L == "scalar":
        map.set_prior(L=1)
    elif L == "vector":
        map.set_prior(L=np.ones(map.Ny))
    elif L == "matrix":
        map.set_prior(L=np.eye(map.Ny))
    elif L == "cholesky":
        map.set_prior(cho_L=np.eye(map.Ny))

    # Provide the dataset
    if C == "scalar":
        map.set_data(flux, C=sigma ** 2)
    elif C == "vector":
        map.set_data(flux, C=np.ones(len(flux)) * sigma ** 2)
    elif C == "matrix":
        map.set_data(flux, C=np.eye(len(flux)) * sigma ** 2)
    elif C == "cholesky":
        map.set_data(flux, cho_C=np.eye(len(flux)) * sigma)

    # Solve the linear problem
    map.inc = inc_true
    mu, cho_cov = map.solve(**kwargs)
    mu = mu.eval()
    cho_cov = cho_cov.eval()

    # Ensure the likelihood of the true value is close to that of
    # the MAP solution
    cov = cho_cov.dot(cho_cov.T)
    LnL0 = multivariate_normal.logpdf(mu, mean=mu, cov=cov)
    LnL = multivariate_normal.logpdf(amp_true * y_true, mean=mu, cov=cov)
    assert LnL0 - LnL < 5.00

    # Check that we can draw from the posterior
    map.draw()


@pytest.mark.parametrize("L,C,woodbury", lnlike_inputs)
def test_lnlike(L, C, woodbury):
    """Test the log marginal likelihood method."""
    # Place a generous prior on the map coefficients
    if L == "scalar":
        map.set_prior(L=1)
    elif L == "vector":
        map.set_prior(L=np.ones(map.Ny))
    elif L == "matrix":
        map.set_prior(L=np.eye(map.Ny))
    elif L == "cholesky":
        map.set_prior(cho_L=np.eye(map.Ny))

    # Provide the dataset
    if C == "scalar":
        map.set_data(flux, C=sigma ** 2)
    elif C == "vector":
        map.set_data(flux, C=np.ones(len(flux)) * sigma ** 2)
    elif C == "matrix":
        map.set_data(flux, C=np.eye(len(flux)) * sigma ** 2)
    elif C == "cholesky":
        map.set_data(flux, cho_C=np.eye(len(flux)) * sigma)

    # Compute the marginal log likelihood for different inclinations
    incs = [15, 30, 45, 60, 75, 90]
    ll = np.zeros_like(incs, dtype=float)
    for i, inc in enumerate(incs):
        map.inc = inc
        ll[i] = map.lnlike(woodbury=woodbury, **kwargs).eval()

    # Verify that we get the correct inclination
    assert incs[np.argmax(ll)] == 60
    assert np.allclose(ll[np.argmax(ll)], 974.221605)  # benchmarked



import numpy as np
import starry
import matplotlib.pyplot as plt
from datetime import datetime
import pytest
from scipy.interpolate import interp1d
from tqdm import tqdm


@pytest.mark.parametrize(
    "xs,ys,zs,source_npts",
    [
        [0, 1, 1, 1],
        [-1, 0, 1, 1],
        [0.5, 1, -0.5, 1],
        [-0.5, -0.5, -0.5, 1],
        [0.5, -0.5, 0.5, 1],
        [1e-08, 0, 1, 1],  # almost noon
        [0, 0, 1, 1],  # exactly noon
        [0, 1, 1, 300],
    ],
)
def test_X(
    xs,
    ys,
    zs,
    source_npts,
    theta=0,
    ro=0.1,
    res=300,
    ydeg=2,
    tol=1e-3,
    plot=False,
):
    # Params
    npts = 250
    xo = np.linspace(-1.5, 1.5, npts)
    yo = np.linspace(-0.3, 0.5, npts)
    theta = 0
    ro = 0.1
    res = 300
    ydeg = 2
    tol = 1e-3

    # Instantiate
    map = starry.Map(ydeg=ydeg, reflected=True, source_npts=source_npts)

    # Analytic
    X = map.amp * map.design_matrix(
        xs=xs, ys=ys, zs=zs, theta=theta, xo=xo, yo=yo, ro=ro
    )

    # Numerical
    (lat, lon), (x, y, z) = map.ops.compute_ortho_grid(res)
    image = np.zeros((map.Ny, res * res))
    image[0] = map.render(theta=theta, xs=xs, ys=ys, zs=zs, res=res).flatten()
    n = 1
    for l in range(1, map.ydeg + 1):
        for m in range(-l, l + 1):
            map.reset()
            map[l, m] = 1
            image[n] = (
                map.render(theta=theta, xs=xs, ys=ys, zs=zs, res=res).flatten()
            ) - image[0]
            n += 1
    X_num = np.zeros_like(X)
    for k in range(len(xo)):
        idx = (x - xo[k]) ** 2 + (y - yo[k]) ** 2 > ro ** 2
        for n in range(map.Ny):
            X_num[k, n] = np.nansum(image[n][idx])
    X_num *= 4 / res ** 2

    # Plot
    if plot:

        fig, ax = plt.subplots(
            ydeg + 1, 2 * ydeg + 1, figsize=(9, 6), sharex=True, sharey=True
        )
        for axis in ax.flatten():
            axis.set_xticks([])
            axis.set_yticks([])
            axis.spines["top"].set_visible(False)
            axis.spines["right"].set_visible(False)
            axis.spines["bottom"].set_visible(False)
            axis.spines["left"].set_visible(False)
        n = 0
        for i, l in enumerate(range(ydeg + 1)):
            for j, m in enumerate(range(-l, l + 1)):
                j += ydeg - l
                med = np.median(X_num[:, n])
                ax[i, j].plot(X[:, n] - med, lw=2)
                ax[i, j].plot(X_num[:, n] - med, lw=1)
                n += 1

        fig.savefig(
            "test_X_{}.pdf".format(datetime.now().strftime("%d%m%Y%H%M%S")),
            bbox_inches="tight",
        )
        plt.close()

    # Compare
    diff = (X - X_num).flatten()
    assert np.max(np.abs(diff)) < tol


def test_inference():
    """
    Test inference on a problem with phase curve + occultations in reflected light.
    The orbital parameters here are contrived to ensure there's no null space;
    note the tiny observational uncertainty as well. Given this setup, a posterior
    map draw should look *very* similar to the true map.

    """
    # Orbital/geometric parameters
    npts = 50000
    t = np.linspace(0, 1, npts)
    porb = 0.19
    prot = 0.12
    rorb = 50
    ro = 38.0
    yo = np.sin(2 * np.pi / porb * t + 0.5)
    xo = np.cos(2 * np.pi / porb * t)
    zo = np.sin(2 * np.pi / porb * t)
    amp = rorb / np.sqrt(xo ** 2 + yo ** 2 + zo ** 2)
    xo *= amp
    yo *= amp
    zo *= amp
    theta = 360.0 / prot * t
    xs = np.sin(7 * np.pi * t)
    ys = np.cos(5 * np.pi * t)
    zs = 5
    kwargs = dict(xs=xs, ys=ys, zs=zs, theta=theta, xo=xo, yo=yo, zo=zo, ro=ro)

    # Generate a synthetic dataset
    map = starry.Map(ydeg=10, reflected=True)
    map.load("earth")
    img0 = map.render(projection="rect", illuminate=False)
    flux0 = map.flux(**kwargs)
    err = 1e-9
    np.random.seed(3)
    flux = flux0 + np.random.randn(npts) * err

    # Solve the linear problem & draw a sample
    map.set_data(flux, C=err ** 2)
    map.set_prior(L=1e-4)
    map.solve(**kwargs)
    map.draw()
    img = map.render(projection="rect", illuminate=False)

    # Verify we recovered the map
    assert np.allclose(img, img0, atol=1e-4)


@pytest.mark.parametrize(
    "b,theta,ro",
    [
        [0.25, np.pi / 3, 0.3],
        [-0.25, np.pi / 3, 0.3],
        [0.25, -np.pi / 3, 0.3],
        [-0.25, -np.pi / 3, 0.3],
        [0.25, 2 * np.pi / 3, 0.3],
        [-0.25, 2 * np.pi / 3, 0.3],
        [0.25, 4 * np.pi / 3, 0.3],
        [-0.25, 4 * np.pi / 3, 0.3],
        [0.5, np.pi / 2, 1.0],
        [0.0, 0.0, 0.5],
        [0.5, 0.0, 0.1],
        [1.0 - 1e-3, 0.0, 0.5],
        [-1.0 + 1e-3, 0.0, 0.5],
        [-1.0, 0.0, 0.5],
        [1.0, 0.0, 0.5],
        [0.25, np.pi / 2, 0.5],
    ],
)
def test_lightcurve(b, theta, ro, ydeg=1, ns=1000, nb=50, res=999, plot=False):

    # Array over full occultation, including all singularities
    xo = 0.0
    yo = np.linspace(0, 1 + ro, ns, endpoint=True)
    for pt in [ro, 1, 1 - ro, b + ro]:
        if pt >= 0:
            yo[np.argmin(np.abs(yo - pt))] = pt
    if theta == 0:
        xs = 0
        ys = 1
    else:
        xs = 0.5
        ys = -xs / np.tan(theta)
    rxy2 = xs ** 2 + ys ** 2
    if b == 0:
        zs = 0
    elif b == 1:
        zs = -1
        xs = 0
        ys = 0
    elif b == -1:
        zs = 1
        xs = 0
        ys = 0
    else:
        zs = -np.sign(b) * np.sqrt(rxy2 / (b ** -2 - 1))

    # Compute analytic
    map = starry.Map(ydeg=ydeg, reflected=True)
    map[1:, :] = 1
    flux = map.flux(xs=xs, ys=ys, zs=zs, xo=xo, yo=yo, ro=ro)

    # Compute numerical
    flux_num = np.zeros_like(yo) * np.nan
    computed = np.zeros(ns, dtype=bool)
    (lat, lon), (x, y, z) = map.ops.compute_ortho_grid(res)
    img = map.render(xs=xs, ys=ys, zs=zs, res=res).flatten()
    for i, yoi in tqdm(enumerate(yo), total=len(yo)):
        if (i == 0) or (i == ns - 1) or (i % (ns // nb) == 0):
            idx = (x - xo) ** 2 + (y - yoi) ** 2 > ro ** 2
            flux_num[i] = np.nansum(img[idx]) * 4 / res ** 2
            computed[i] = True

    # Interpolate over numerical result
    f = interp1d(yo[computed], flux_num[computed], kind="cubic")
    flux_num_interp = f(yo)

    # Plot
    if plot:
        fig = plt.figure()
        plt.plot(yo, flux, "C0-", label="starry", lw=2)
        plt.plot(yo, flux_num, "C1o", label="brute")
        plt.plot(yo, flux_num_interp, "C1-", lw=1)
        plt.legend(loc="best")
        plt.xlabel("impact parameter")
        plt.ylabel("flux")
        fig.savefig(
            "test_lightcurve[{}-{}-{}].pdf".format(b, theta, ro),
            bbox_inches="tight",
        )
        plt.close()

    # Compare with very lax tolerance; we're mostly looking
    # for gross outliers
    diff = np.abs(flux - flux_num_interp)
    assert np.max(diff) < 0.001


@pytest.mark.parametrize(
    "b,theta,bo,ro",
    [
        #
        # Occultor does not touch the terminator
        #
        [0.5, 0.1, 1.2, 0.1],
        [0.5, 0.1, 0.1, 1.2],
        [0.5, 0.1, 0.8, 0.1],
        [0.5, 0.1, 0.9, 0.2],
        [0.5, np.pi + 0.1, 0.8, 0.1],
        [0.5, np.pi + 0.1, 0.9, 0.2],
        [0.5, 0.1, 0.5, 1.25],
        [0.5, np.pi + 0.1, 0.5, 1.25],
        #
        # Occultations involving all three primitive integrals
        #
        [0.4, np.pi / 3, 0.5, 0.7],
        [0.4, 2 * np.pi - np.pi / 3, 0.5, 0.7],
        [0.4, np.pi / 2, 0.5, 0.7],
        [0.4, np.pi / 2, 1.0, 0.2],
        [0.00001, np.pi / 2, 0.5, 0.7],
        [0, np.pi / 2, 0.5, 0.7],
        [0.4, -np.pi / 2, 0.5, 0.7],
        [-0.4, np.pi / 3, 0.5, 0.7],
        [-0.4, 2 * np.pi - np.pi / 3, 0.5, 0.7],
        [-0.4, np.pi / 2, 0.5, 0.7],
        #
        # Occultations involving only P and T
        #
        [0.4, np.pi / 6, 0.3, 0.3],
        [0.4, np.pi + np.pi / 6, 0.1, 0.6],
        [0.4, np.pi + np.pi / 3, 0.1, 0.6],
        [0.4, np.pi / 6, 0.6, 0.5],
        [0.4, -np.pi / 6, 0.6, 0.5],
        [0.4, 0.1, 2.2, 2.0],
        [0.4, -0.1, 2.2, 2.0],
        [0.4, np.pi + np.pi / 6, 0.3, 0.8],
        [0.75, np.pi + 0.1, 4.5, 5.0],
        [-0.95, 0.0, 2.0, 2.5],
        [-0.1, np.pi / 6, 0.6, 0.75],
        [-0.5, np.pi, 0.8, 0.5],
        [-0.1, 0.0, 0.5, 1.0],
        #
        # Occultations involving three points of intersection with the terminator
        #
        [
            0.5488316824842527,
            4.03591586925189,
            0.34988513192814663,
            0.7753986686719786,
        ],
        [
            0.5488316824842527,
            2 * np.pi - 4.03591586925189,
            0.34988513192814663,
            0.7753986686719786,
        ],
        [
            -0.5488316824842527,
            4.03591586925189 - np.pi,
            0.34988513192814663,
            0.7753986686719786,
        ],
        [
            -0.5488316824842527,
            2 * np.pi - (4.03591586925189 - np.pi),
            0.34988513192814663,
            0.7753986686719786,
        ],
        #
        # Occultations involving four points of intersection with the terminator
        #
        [0.5, np.pi, 0.99, 1.5],
        [-0.5, 0.0, 0.99, 1.5],
        #
        # Miscellaneous edge cases
        #
        [0.5, np.pi, 1.0, 1.5],
        [0.5, 2 * np.pi - np.pi / 4, 0.4, 0.4],
        [0.5, 2 * np.pi - np.pi / 4, 0.3, 0.3],
        [-0.25, 4 * np.pi / 3, 0.3, 0.3],
    ],
)
def test_cases(b, theta, bo, ro, ydeg=1, res=999):

    # Array over full occultation, including all singularities
    xo = 0.0
    yo = bo
    if theta == 0:
        xs = 0
        ys = 1
    else:
        xs = 0.5
        ys = -xs / np.tan(theta)
    rxy2 = xs ** 2 + ys ** 2
    if b == 0:
        zs = 0
    elif b == 1:
        zs = -1
        xs = 0
        ys = 0
    elif b == -1:
        zs = 1
        xs = 0
        ys = 0
    else:
        zs = -np.sign(b) * np.sqrt(rxy2 / (b ** -2 - 1))

    # Compute analytic
    map = starry.Map(ydeg=ydeg, reflected=True)
    map[1:, :] = 1
    flux = map.flux(xs=xs, ys=ys, zs=zs, xo=xo, yo=yo, ro=ro)

    # Compute numerical
    (lat, lon), (x, y, z) = map.ops.compute_ortho_grid(res)
    img = map.render(xs=xs, ys=ys, zs=zs, res=res).flatten()
    idx = (x - xo) ** 2 + (y - yo) ** 2 > ro ** 2
    flux_num = np.nansum(img[idx]) * 4 / res ** 2

    # Compare with very lax tolerance; we're mostly looking
    # for gross outliers
    diff = np.abs(flux - flux_num)
    assert diff < 0.001


def test_theta_poles(res=500, tol=1e-3):
    """Test cases near the poles for theta."""
    # Settings
    ydeg = 10
    zs = -0.25
    xo = 0.0
    yo = 0.35
    ro = 0.25
    n = 5

    # Compare
    map = starry.Map(ydeg, reflected=True)
    map[ydeg, :] = 1
    x = np.array([0.0, 0.5, 1.0, 1.5, 2.0]).reshape(-1, 1) * np.pi
    dx = np.concatenate(
        (-np.logspace(-15, -5, n)[::-1], [0], np.logspace(-15, -5, n))
    ).reshape(1, -1)
    theta = (x + dx).reshape(-1)
    (lat, lon), (x, y, z) = map.ops.compute_ortho_grid(res)
    err = np.zeros_like(theta)
    for i in range(len(theta)):
        if theta[i] == 0:
            xs = 0
            ys = 1
        else:
            xs = 0.5
            ys = -xs / np.tan(theta[i])
        flux = map.flux(xs=xs, ys=ys, zs=zs, xo=xo, yo=yo, ro=ro)
        img = map.render(xs=xs, ys=ys, zs=zs, res=res).flatten()
        idx = (x - xo) ** 2 + (y - yo) ** 2 > ro ** 2
        flux_num = np.nansum(img[idx]) * 4 / res ** 2
        err[i] = np.max(np.abs(flux - flux_num))
    assert np.all(err < tol)


# BROKEN: Figure out why the root finder fails here.
@pytest.mark.xfail
def test_root_finder():
    """
    Test cases that cause the root finder to fail.

    """
    map = starry.Map(reflected=True)
    map.ops._sT.func([-0.358413], [-1.57303], [55.7963], 54.8581, 0.0)


# BROKEN: Figure this out
@pytest.mark.xfail
def test_bad_case():
    """
    Test pathological wrong case identification.

    """
    map = starry.Map(reflected=True)

    # These values lead to a (very) wrong flux
    theta0 = -0.0409517311212404
    b0 = -0.83208413089546
    bo0 = 12.073565287605442
    ro = 12.155639360414618

    # Perturb theta in the vicinity of theta0
    delta = np.linspace(0, 1e-6, 100)
    theta = np.concatenate((theta0 - delta[::-1], theta0 + delta))

    # Compute the flux
    b = b0 * np.ones_like(theta)
    bo = bo0 * np.ones_like(theta)
    sT, *_ = map.ops._sT.func(b, theta, bo, ro, 0.0)
    flux = sT[:, 0]

    # DEBUG
    # plt.plot(theta, flux)
    # plt.show()

    # Check that it's approximately constant over the range
    assert np.allclose(flux, flux[0])



# -*- coding: utf-8 -*-
"""Test adding spots to the map."""
import starry
import numpy as np
from scipy.ndimage import gaussian_filter1d


def test_normalization():
    map = starry.Map(30)
    contrast = 0.75
    baseline = 1 / np.pi  # the default starry intensity baseline
    map.spot(contrast=contrast, radius=20, lat=0, lon=0)
    actual_contrast = (baseline - map.intensity(lat=0, lon=0)) / baseline
    assert np.abs(contrast - actual_contrast) < 0.01


def test_profile():
    # Setttings (hand-tuned)
    radius = 20
    spot_fac = 300
    spot_smoothing = 0.075
    gaussian_smoothing = 25
    med_tol = 0.01
    max_tol = 0.05

    # Get the starry profile
    map = starry.Map(30)
    map.spot(
        contrast=1,
        radius=radius,
        lat=0,
        lon=0,
        spot_fac=spot_fac,
        spot_smoothing=spot_smoothing,
    )
    lon = np.linspace(-90, 90, 1000)
    I = np.pi * map.intensity(lon=lon)

    # The exact sigmoid we're expanding, with a Gaussian smoothing filter
    I0 = 1.0 / (1.0 + np.exp(-spot_fac * (np.abs(lon) - radius)))
    I0 = gaussian_filter1d(I0, gaussian_smoothing)

    # Check that they agree
    assert np.median(np.abs(I - I0)) < med_tol
    assert np.max(np.abs(I - I0)) < max_tol


def test_latlon():
    lat0 = 30
    lon0 = 80
    map = starry.Map(20)
    map.spot(lat=lat0, lon=lon0)
    lat = np.arange(-90, 90, dtype="float64")
    lon = np.arange(-180, 180, dtype="float64")
    lat, lon = np.meshgrid(lat, lon)
    lat = lat.reshape(-1)
    lon = lon.reshape(-1)
    I = map.intensity(lat=lat, lon=lon)
    imin = np.argmin(I)
    assert lat[imin] == lat0
    assert lon[imin] == lon0


def test_two_spots():
    """
    Test the behavior when adding two identical spots in different locations.
    The intensities at the spot centers should be the same.

    """

    map = starry.Map(ydeg=10)
    map.spot(lon=0)
    map.spot(lon=90)
    return np.isclose(map.intensity(lon=0), map.intensity(lon=90))



# -*- coding: utf-8 -*-
from .. import config
from ..compat import Node, change_flags, theano, tt
import numpy as np
from inspect import getmro
from functools import wraps
import logging
import sys

logger = logging.getLogger("starry.ops")

__all__ = ["logger", "autocompile", "is_tensor", "clear_cache"]


integers = (int, np.int16, np.int32, np.int64)


def is_tensor(*objs):
    """Return ``True`` if any of ``objs`` is a ``Theano`` object."""
    for obj in objs:
        for c in getmro(type(obj)):
            if c is Node:
                return True
    return False


class CompileLogMessage:
    """
    Log a brief message saying what method is currently
    being compiled and print `Done` when finished.

    """

    def __init__(self, name):
        self.name = name

    def __enter__(self):
        config.rootHandler.terminator = ""
        logger.info("Compiling `{0}`... ".format(self.name))

    def __exit__(self, type, value, traceback):
        config.rootHandler.terminator = "\n"
        logger.info("Done.")


def _get_type(arg):
    """
    Get the theano tensor type corresponding to `arg`.

    Note that arg must be one of the following:
        - a theano tensor
        - an integer (`int`, `np.int`, `np.int16`, `np.int32`, `np.int64`)
        - a numpy boolean (`np.array(True)`, `np.array(False)`)
        - a numpy float array with ndim equal to 0, 1, 2, or 3

    # TODO: Cast lists to arrays and floats to np.array(float)

    """
    ttype = type(arg)
    if is_tensor(arg):
        return ttype
    else:
        if ttype in integers:
            return tt.iscalar
        elif hasattr(arg, "ndim"):
            if arg.ndim == 0:
                if arg.dtype is np.array(True).dtype:
                    return tt.bscalar
                else:
                    return tt.dscalar
            elif arg.ndim == 1:
                return tt.dvector
            elif arg.ndim == 2:
                return tt.dmatrix
            elif arg.ndim == 3:
                return tt.dtensor3
            else:
                raise NotImplementedError(
                    "Invalid array dimension passed to @autocompile: {}.".format(
                        arg.ndim
                    )
                )
        else:
            raise NotImplementedError(
                "Invalid argument type passed to @autocompile: {}.".format(
                    ttype
                )
            )


def autocompile(func):
    """
    Wrap the method `func` and return a compiled version
    if none of the arguments are tensors.

    """

    @wraps(func)  # inherit docstring
    def wrapper(instance, *args):

        if is_tensor(*args):

            # Just return the function as is
            return func(instance, *args)

        else:

            # Determine the argument types
            arg_types = tuple([_get_type(arg) for arg in args])

            # Get a unique name for the compiled function
            cname = "__{}_{}".format(
                func.__name__, hex(hash(arg_types) % ((sys.maxsize + 1) * 2))
            )

            # Compile the function if needed & cache it
            if not hasattr(instance, cname):

                dummy_args = [arg_type() for arg_type in arg_types]

                # Compile the function
                with CompileLogMessage(func.__name__):
                    with change_flags(compute_test_value="off"):
                        compiled_func = theano.function(
                            [*dummy_args],
                            func(instance, *dummy_args),
                            on_unused_input="ignore",
                            profile=config.profile,
                            mode=config.mode,
                        )
                    setattr(instance, cname, compiled_func)

            # Return the compiled version
            return getattr(instance, cname)(*args)

    return wrapper


def clear_cache(instance, func):
    """
    Clear the compiled function cache for method `func` of a class
    instance `instance`.

    """
    basename = "__{}_".format(func.__name__)
    for key in list(instance.__dict__.keys()):
        if key.startswith(basename):
            delattr(instance, key)



import starry
from starry.compat import theano
from starry.compat import tt
from starry.compat import change_flags
import pytest
import numpy as np


def test_zero_lat_lon(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    """
    The gradient of the intensity is NaN when either the
    latitude or the longitude is zero. There's a div by zero in
    the Op that we need to work around.

    """

    lats = [0.0001, 0.0, np.pi / 6, 0.0]
    lons = [np.pi / 6, np.pi / 6, 0.0, 0.0]

    for lat, lon in zip(lats, lons):
        with change_flags(compute_test_value="off"):
            map = starry.Map(ydeg=2, udeg=2)
            np.random.seed(11)
            y = np.array([1.0] + list(np.random.randn(8)))
            u = np.array([-1.0] + list(np.random.randn(2)))
            f = np.array([np.pi])
            theta = np.array(0.0)
            lat = np.array(lat)
            lon = np.array(lon)

            def intensity(lat, lon, y, u, f, theta):
                return map.ops.intensity(
                    lat, lon, y, u, f, theta, np.array(True)
                )

            theano.gradient.verify_grad(
                intensity,
                (lat, lon, y, u, f, theta),
                abs_tol=abs_tol,
                rel_tol=rel_tol,
                eps=eps,
                n_tests=1,
                rng=np.random,
            )



# -*- coding: utf-8 -*-
"""
Test against numerical integration.

"""
from starry.compat import theano
import starry
import numpy as np

res = 1000
map_tmp = starry.Map(1)


def flux(map, xs=0, ys=0, zs=0, xo=0, yo=0, zo=1, ro=0, **kwargs):
    (lat, lon), (x, y, z) = map_tmp.ops.compute_ortho_grid(res)
    occulted = (xo - x) ** 2 + (yo - y) ** 2 < ro ** 2
    occulted = occulted.reshape(res, res)
    if map._ops_class_ == starry._core.OpsReflected:
        image = map.render(res=res, xs=xs, ys=ys, zs=zs, **kwargs)
    else:
        image = map.render(res=res, **kwargs)
    image[occulted] = 0.0
    image *= 4.0 / (res ** 2)
    return np.nansum(image)


def test_flux(n_tests=10):
    map = starry.Map(2)
    np.random.seed(12)
    for i in range(n_tests):
        map[1:, :] = np.random.randn(len(map[1:, :]))
        theta = np.random.random() * 360
        xo = np.random.randn()
        yo = np.random.randn()
        ro = 0.5 * np.random.random()
        kwargs = dict(theta=theta, xo=xo, yo=yo, ro=ro)
        assert np.allclose(
            map.flux(**kwargs), flux(map, **kwargs), rtol=1e-3, atol=1e-3
        )


def test_flux_reflected(n_tests=10):
    map = starry.Map(2, reflected=True)
    np.random.seed(13)
    for i in range(n_tests):
        map[1:, :] = 0.1 * np.random.randn(len(map[1:, :]))
        theta = np.random.random() * 360
        source = np.random.randn(3)
        kwargs = dict(
            theta=theta, xs=source[0], ys=source[1], zs=source[2], ro=0
        )
        assert np.allclose(
            map.flux(**kwargs), flux(map, **kwargs), rtol=1e-3, atol=1e-3
        )


if __name__ == "__main__":
    test_flux_reflected()



# -*- coding: utf-8 -*-
"""Test Keplerian system stuff."""
import starry
import pytest
from starry.compat import theano
import numpy as np


def test_orientation(Omega=45, inc=35):
    # Instantiate
    pri = starry.Primary(starry.Map(amp=0))
    sec = starry.Secondary(
        starry.Map(ydeg=1, amp=1),
        porb=1.0,
        r=0,
        m=0,
        inc=inc,
        Omega=Omega,
        prot=1.0,
        theta0=180.0,
    )
    sec.map[1, 0] = 1.0
    sys = starry.System(pri, sec)

    # Align the rotational axis with the orbital axis
    sec.map.inc = sec.inc
    sec.map.obl = sec.Omega

    # Compute the flux
    t = np.linspace(-0.5, 0.5, 1000)
    flux = sys.flux(t)

    # This is the analytic result
    flux_analytic = 1.0 - np.sin(inc * np.pi / 180.0) * (
        2.0 / np.sqrt(3.0)
    ) * np.cos(2 * np.pi * t)

    assert np.allclose(flux, flux_analytic)


def test_bodies():
    pri = starry.Primary(starry.Map())
    sec = starry.Secondary(starry.Map(ydeg=1), porb=1.0)
    sys = starry.System(pri, sec)
    assert sys.primary == pri
    assert sys.secondaries[0] == sec


def test_integration():
    pri = starry.Primary(starry.Map(udeg=2), r=1.0)
    pri.map[1:] = [0.5, 0.25]
    sec = starry.Secondary(starry.Map(ydeg=1), porb=1.0, r=0.25)

    # Manual integration
    t = np.linspace(-0.1, 0.1, 10000)
    sys = starry.System(pri, sec, texp=0)
    flux = sys.flux(t)
    t = t.reshape(-1, 1000).mean(axis=1)
    flux = flux.reshape(-1, 1000).mean(axis=1)

    sys0 = starry.System(pri, sec, texp=0.02, order=0, oversample=999)
    assert sys0.order == 0
    assert sys0.oversample == 999
    flux0 = sys0.flux(t)
    assert np.allclose(flux, flux0)

    sys1 = starry.System(pri, sec, texp=0.02, order=1, oversample=999)
    assert sys1.order == 1
    assert sys1.oversample == 999
    flux1 = sys1.flux(t)
    assert np.allclose(flux, flux1)

    sys2 = starry.System(pri, sec, texp=0.02, order=2, oversample=999)
    assert sys2.order == 2
    assert sys2.oversample == 999
    flux2 = sys2.flux(t)
    assert np.allclose(flux, flux2)


def test_reflected_light():
    pri = starry.Primary(starry.Map(amp=0), r=1)
    sec = starry.Secondary(starry.Map(reflected=True), porb=1.0, r=1)
    sys = starry.System(pri, sec)
    t = np.concatenate((np.linspace(0.1, 0.4, 50), np.linspace(0.6, 0.9, 50)))
    flux = sys.flux(t)

    # TODO: Add an analytic validation here



# -*- coding: utf-8 -*-
from .. import config
from ..compat import tt, slinalg, floatX
from .._constants import *
from .utils import *
import numpy as np
from scipy.linalg import block_diag as scipy_block_diag
import scipy

__all__ = ["lazy_math", "greedy_math", "lazy_linalg", "greedy_linalg"]


# Cholesky solve
_solve_lower = slinalg.Solve(A_structure="lower_triangular", lower=True)
_solve_upper = slinalg.Solve(A_structure="upper_triangular", lower=False)


def _cho_solve(cho_A, b):
    return _solve_upper(tt.transpose(cho_A), _solve_lower(cho_A, b))


def _get_covariance(math, linalg, C=None, cho_C=None, N=None):
    """A container for covariance matrices.

    Args:
        C (scalar, vector, or matrix, optional): The covariance.
            Defaults to None.
        cho_C (matrix, optional): The lower Cholesky factorization of
            the covariance. Defaults to None.
        N (int, optional): The number of rows/columns in the covariance
            matrix, required if ``C`` is a scalar. Defaults to None.
    """

    # User provided the Cholesky factorization
    if cho_C is not None:

        cholesky = math.cast(cho_C)
        value = math.dot(cholesky, math.transpose(cholesky))
        inverse = linalg.cho_solve(cholesky, math.eye(cholesky.shape[0]))
        lndet = 2 * math.sum(math.log(math.diag(cholesky)))
        kind = "cholesky"
        N = cho_C.shape[0]

    # User provided the covariance as a scalar, vector, or matrix
    elif C is not None:

        C = math.cast(C)

        if hasattr(C, "ndim"):

            if C.ndim == 0:

                assert N is not None, "Please provide a matrix size `N`."
                cholesky = math.sqrt(C)
                inverse = math.cast(1.0 / C)
                lndet = math.cast(N * math.log(C))
                value = C
                kind = "scalar"

            elif C.ndim == 1:

                cholesky = math.sqrt(C)
                inverse = 1.0 / C
                lndet = math.sum(math.log(C))
                value = C
                kind = "vector"
                N = C.shape[0]

            else:

                cholesky = math.cholesky(C)
                inverse = linalg.cho_solve(cholesky, math.eye(C.shape[0]))
                lndet = 2 * math.sum(math.log(math.diag(cholesky)))
                value = C
                kind = "matrix"
                N = C.shape[0]

        # Assume it's a scalar
        else:

            assert N is not None, "Please provide a matrix size `N`."
            cholesky = math.sqrt(C)
            inverse = math.cast(1.0 / C)
            lndet = math.cast(N * math.log(C))
            value = C
            kind = "scalar"

    # ?!
    else:
        raise ValueError(
            "Either the covariance or its Cholesky factorization must be provided."
        )

    return value, cholesky, inverse, lndet, kind, N


class MathType(type):
    """Wrapper for theano/numpy functions."""

    def cholesky(cls, *args, **kwargs):
        if cls.lazy:
            return slinalg.cholesky(*args, **kwargs)
        else:
            return scipy.linalg.cholesky(*args, **kwargs, lower=True)

    def atleast_2d(cls, arg):
        if cls.lazy:
            return arg * tt.ones((1, 1))
        else:
            return np.atleast_2d(arg)

    def vectorize(cls, *args):
        """
        Vectorize all ``args`` so that they have the same length
        along the first axis.

        TODO: Add error catching if the dimensions don't agree.
        """
        if cls.lazy:
            args = [arg * tt.ones(1) for arg in args]
            size = tt.max([arg.shape[0] for arg in args])
            args = [tt.repeat(arg, size // arg.shape[0], 0) for arg in args]
        else:
            args = [np.atleast_1d(arg) for arg in args]
            size = np.max([arg.shape[0] for arg in args])
            args = tuple(
                [
                    arg
                    * np.ones(
                        (size,) + tuple(np.ones(len(arg.shape) - 1, dtype=int))
                    )
                    for arg in args
                ]
            )
        if len(args) == 1:
            return args[0]
        else:
            return args

    def cross(cls, x, y):
        """Cross product of two 3-vectors.

        Based on ``https://github.com/Theano/Theano/pull/3008``
        """
        if cls.lazy:
            eijk = np.zeros((3, 3, 3))
            eijk[0, 1, 2] = eijk[1, 2, 0] = eijk[2, 0, 1] = 1
            eijk[0, 2, 1] = eijk[2, 1, 0] = eijk[1, 0, 2] = -1
            return tt.as_tensor_variable(tt.dot(tt.dot(eijk, y), x))
        else:
            return np.cross(x, y)

    def cast(cls, *args):
        if cls.lazy:
            return cls.to_tensor(*args)
        else:
            if len(args) == 1:
                return np.array(args[0], dtype=floatX)
            else:
                return [np.array(arg, dtype=floatX) for arg in args]

    def to_array_or_tensor(cls, x):
        if cls.lazy:
            return tt.as_tensor_variable(x)
        else:
            return np.array(x)

    def block_diag(cls, *mats):
        if cls.lazy:
            N = [mat.shape[0] for mat in mats]
            Nsum = tt.sum(N)
            res = tt.zeros((Nsum, Nsum), dtype=floatX)
            n = 0
            for mat in mats:
                inds = slice(n, n + mat.shape[0])
                res = tt.set_subtensor(res[tuple((inds, inds))], mat)
                n += mat.shape[0]
            return res
        else:
            return scipy_block_diag(*mats)

    def to_tensor(cls, *args):
        """Convert all ``args`` to Theano tensor variables.

        Converts to tensor regardless of whether `cls.lazy` is True or False.
        """
        if len(args) == 1:
            return tt.as_tensor_variable(args[0]).astype(floatX)
        else:
            return [tt.as_tensor_variable(arg).astype(floatX) for arg in args]

    def __getattr__(cls, attr):
        if cls.lazy:
            return getattr(tt, attr)
        else:
            return getattr(np, attr)


class LinAlgType(type):
    """Linear algebra operations."""

    @autocompile
    def cho_solve(self, cho_A, b):
        return _cho_solve(cho_A, b)

    @autocompile
    def solve(self, X, flux, cho_C, mu, LInv):
        """
        Compute the maximum a posteriori (MAP) prediction for the
        spherical harmonic coefficients of a map given a flux timeseries.

        Args:
            X (matrix): The flux design matrix.
            flux (array): The flux timeseries.
            cho_C (scalar/vector/matrix): The lower cholesky factorization
                of the data covariance.
            mu (array): The prior mean of the spherical harmonic coefficients.
            LInv (scalar/vector/matrix): The inverse prior covariance of the
                spherical harmonic coefficients.

        Returns:
            The vector of spherical harmonic coefficients corresponding to the
            MAP solution and the Cholesky factorization of the corresponding
            covariance matrix.

        """
        # TODO: These if statements won't play well with @autocompile!!!

        # Compute C^-1 . X
        if cho_C.ndim == 0:
            CInvX = X / cho_C ** 2
        elif cho_C.ndim == 1:
            CInvX = tt.dot(tt.diag(1 / cho_C ** 2), X)
        else:
            CInvX = _cho_solve(cho_C, X)

        # Compute W = X^T . C^-1 . X + L^-1
        W = tt.dot(tt.transpose(X), CInvX)
        if LInv.ndim == 0:
            W = tt.inc_subtensor(
                W[tuple((tt.arange(W.shape[0]), tt.arange(W.shape[0])))], LInv
            )
            LInvmu = mu * LInv
        elif LInv.ndim == 1:
            W = tt.inc_subtensor(
                W[tuple((tt.arange(W.shape[0]), tt.arange(W.shape[0])))], LInv
            )
            LInvmu = mu * LInv
        else:
            W += LInv
            LInvmu = tt.dot(LInv, mu)

        # Compute the max like y and its covariance matrix
        cho_W = slinalg.cholesky(W)
        M = _cho_solve(cho_W, tt.transpose(CInvX))
        yhat = tt.dot(M, flux) + _cho_solve(cho_W, LInvmu)
        ycov = _cho_solve(cho_W, tt.eye(cho_W.shape[0]))
        cho_ycov = slinalg.cholesky(ycov)

        return yhat, cho_ycov

    @autocompile
    def lnlike(cls, X, flux, C, mu, L):
        """
        Compute the log marginal likelihood of the data given a design matrix.

        Args:
            X (matrix): The flux design matrix.
            flux (array): The flux timeseries.
            C (scalar/vector/matrix): The data covariance matrix.
            mu (array): The prior mean of the spherical harmonic coefficients.
            L (scalar/vector/matrix): The prior covariance of the spherical
                harmonic coefficients.

        Returns:
            The log marginal likelihood of the `flux` vector conditioned on
            the design matrix `X`. This is the likelihood marginalized over
            all possible spherical harmonic vectors, which is analytically
            computable for the linear `starry` model.

        """
        # TODO: These if statements won't play well with @autocompile!!!

        # Compute the GP mean
        gp_mu = tt.dot(X, mu)

        # Compute the GP covariance
        if L.ndim == 0:
            XLX = tt.dot(X, tt.transpose(X)) * L
        elif L.ndim == 1:
            XLX = tt.dot(tt.dot(X, tt.diag(L)), tt.transpose(X))
        else:
            XLX = tt.dot(tt.dot(X, L), tt.transpose(X))

        if C.ndim == 0 or C.ndim == 1:
            gp_cov = tt.inc_subtensor(
                XLX[tuple((tt.arange(XLX.shape[0]), tt.arange(XLX.shape[0])))],
                C,
            )
        else:
            gp_cov = C + XLX

        cho_gp_cov = slinalg.cholesky(gp_cov)

        # Compute the marginal likelihood
        N = X.shape[0]
        r = tt.reshape(flux - gp_mu, (-1, 1))
        lnlike = -0.5 * tt.dot(tt.transpose(r), _cho_solve(cho_gp_cov, r))
        lnlike -= tt.sum(tt.log(tt.diag(cho_gp_cov)))
        lnlike -= 0.5 * N * tt.log(2 * np.pi)

        return lnlike[0, 0]

    @autocompile
    def lnlike_woodbury(cls, X, flux, CInv, mu, LInv, lndetC, lndetL):
        """
        Compute the log marginal likelihood of the data given a design matrix
        using the Woodbury identity.

        Args:
            X (matrix): The flux design matrix.
            flux (array): The flux timeseries.
            CInv (scalar/vector/matrix): The inverse data covariance matrix.
            mu (array): The prior mean of the spherical harmonic coefficients.
            L (scalar/vector/matrix): The inverse prior covariance of the
                spherical harmonic coefficients.

        Returns:
            The log marginal likelihood of the `flux` vector conditioned on
            the design matrix `X`. This is the likelihood marginalized over
            all possible spherical harmonic vectors, which is analytically
            computable for the linear `starry` model.

        """
        # TODO: These if statements won't play well with @autocompile!!!

        # Compute the GP mean
        gp_mu = tt.dot(X, mu)

        # Residual vector
        r = tt.reshape(flux - gp_mu, (-1, 1))

        # Inverse of GP covariance via Woodbury identity
        if CInv.ndim == 0:
            U = X * CInv
        elif CInv.ndim == 1:
            U = tt.dot(tt.diag(CInv), X)
        else:
            U = tt.dot(CInv, X)

        if LInv.ndim == 0:
            W = tt.dot(tt.transpose(X), U) + LInv * tt.eye(U.shape[1])
        elif LInv.ndim == 1:
            W = tt.dot(tt.transpose(X), U) + tt.diag(LInv)
        else:
            W = tt.dot(tt.transpose(X), U) + LInv
        cho_W = slinalg.cholesky(W)

        if CInv.ndim == 0:
            SInv = CInv * tt.eye(U.shape[0]) - tt.dot(
                U, _cho_solve(cho_W, tt.transpose(U))
            )
        elif CInv.ndim == 1:
            SInv = tt.diag(CInv) - tt.dot(
                U, _cho_solve(cho_W, tt.transpose(U))
            )
        else:
            SInv = CInv - tt.dot(U, _cho_solve(cho_W, tt.transpose(U)))

        # Determinant of GP covariance
        lndetW = 2 * tt.sum(tt.log(tt.diag(cho_W)))
        lndetS = lndetW + lndetC + lndetL

        # Compute the marginal likelihood
        N = X.shape[0]
        lnlike = -0.5 * tt.dot(tt.transpose(r), tt.dot(SInv, r))
        lnlike -= 0.5 * lndetS
        lnlike -= 0.5 * N * tt.log(2 * np.pi)

        return lnlike[0, 0]

    @autocompile
    def _cho_solve(cls, cho_A, b):
        return _cho_solve(cho_A, b)


class lazy_math(metaclass=MathType):
    """Alias for ``numpy`` or ``theano.tensor``."""

    lazy = True


class greedy_math(metaclass=MathType):
    """Alias for ``numpy`` or ``theano.tensor``."""

    lazy = False


class lazy_linalg(metaclass=LinAlgType):
    """Miscellaneous linear algebra operations."""

    lazy = True

    class Covariance(object):
        def __init__(self, *args, **kwargs):
            (
                self.value,
                self.cholesky,
                self.inverse,
                self.lndet,
                self.kind,
                self.N,
            ) = _get_covariance(lazy_math, lazy_linalg, *args, **kwargs)


class greedy_linalg(metaclass=LinAlgType):
    """Miscellaneous linear algebra operations."""

    lazy = False

    class Covariance(object):
        def __init__(self, *args, **kwargs):
            (
                self.value,
                self.cholesky,
                self.inverse,
                self.lndet,
                self.kind,
                self.N,
            ) = _get_covariance(greedy_math, greedy_linalg, *args, **kwargs)



# This file is part of Eigen, a lightweight C++ template library
# for linear algebra.
#
# Copyright (C) 2012 Keir Mierle <mierle@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: mierle@gmail.com (Keir Mierle)
#
# Make the long-awaited conversion to MPL.

lgpl3_header = '''
// Eigen is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 3 of the License, or (at your option) any later version.
//
// Alternatively, you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// Eigen is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License or the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License and a copy of the GNU General Public License along with
// Eigen. If not, see <http://www.gnu.org/licenses/>.
'''

mpl2_header = """
// This Source Code Form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import os
import sys

exclusions = set(['relicense.py'])

def update(text):
  if text.find(lgpl3_header) == -1:
    return text, False
  return text.replace(lgpl3_header, mpl2_header), True

rootdir = sys.argv[1]
for root, sub_folders, files in os.walk(rootdir):
    for basename in files:
        if basename in exclusions:
          print 'SKIPPED', filename
          continue
        filename = os.path.join(root, basename)
        fo = file(filename)
        text = fo.read()
        fo.close()

        text, updated = update(text)
        if updated:
          fo = file(filename, "w")
          fo.write(text)
          fo.close()
          print 'UPDATED', filename
        else:
          print '       ', filename



# -*- coding: utf-8 -*-
"""
Functions to test map importing / SH transforms

"""
import starry
import numpy as np
from PIL import Image
from matplotlib.image import pil_to_array
import os
import pytest
from skimage.feature import canny
from scipy.ndimage import shift


def test_load():
    """Test that we're loading images correctly."""
    # Get the image directly
    image = os.path.join(os.path.dirname(starry.__file__), "img", "spot.png")
    grayscale_pil_image = Image.open(image).convert("L")
    img1 = pil_to_array(grayscale_pil_image)
    img1 = np.array(img1, dtype=float)
    img1 /= 255.0  # to [0, 1] range
    img1 = np.flipud(img1)  # align
    img1 = img1[:, ::2]  # make square to match starry output

    # Render the image with starry at high res
    map = starry.Map(30)
    map.load("spot", extent=(-180, 179, -90, 89))
    img2 = map.render(projection="rect", res=img1.shape[0])

    # Canny filter on both images to detect edges
    img1 = np.array(canny(img1, sigma=4), dtype=int)
    img2 = np.array(canny(img2, sigma=4), dtype=int)

    # Compute the difference between the images
    # Do this several times by translating one of
    # the images by a single pixel and computing
    # the *minimum*. This has the effect of allowing
    # a one-pixel tolerance in the comparison
    diff = np.min(
        np.abs(
            [
                img2 - img1,
                img2 - shift(img1, [0, 1]),
                img2 - shift(img1, [0, -1]),
                img2 - shift(img1, [1, 0]),
                img2 - shift(img1, [-1, 0]),
            ]
        ),
        axis=0,
    )

    # Number of pixels in the image edges
    pixels = np.count_nonzero(img1)

    # Number of pixels that differ between the two
    pixels_diff = np.count_nonzero(diff)

    # There should be VERY few pixels in the difference image
    assert pixels_diff / pixels < 0.01


def test_load_normalization():
    """Test that we're normalizing imported images correctly."""
    # Render a band-limited image
    map = starry.Map(5)
    map[5, 3] = 1
    img1 = map.render(projection="moll")
    img1_rect = map.render(projection="rect")

    # Now load that image and re-render it;
    # the resulting map should be *very*
    # close to the original map!
    map.reset()
    map.load(img1_rect, fac=np.inf, smoothing=0)
    img2 = map.render(projection="moll")

    # Compute stats
    mu1 = np.nanmean(img1)
    sig1 = np.nanstd(img1)
    mu2 = np.nanmean(img2)
    sig2 = np.nanstd(img2)

    # Assert diff is small
    assert np.abs(mu1 - mu2) < 1e-3
    assert np.abs(sig1 - sig2) < 1e-3


def test_force_psd():
    """Test the positive semi-definite feature."""
    # High res map
    map = starry.Map(10)

    # Render the image with starry with aggressive minimum finding
    map.load("spot", force_psd=True, oversample=3, ntries=2)

    # Ensure positive everywhere
    assert map.render(projection="rect").min() >= 0



# -*- coding: utf-8 -*-
from ....compat import COp
from ....starry_version import __version__
import sys
import pkg_resources


__all__ = ["LimbDarkBaseOp"]


class LimbDarkBaseOp(COp):

    __props__ = ()
    func_file = None
    func_name = None

    def __init__(self):
        super(LimbDarkBaseOp, self).__init__(self.func_file, self.func_name)

    def c_code_cache_version(self, *args, **kwargs):
        if "dev" in __version__:
            return ()
        return tuple(map(int, __version__.split(".")))

    def c_headers(self, *args, **kwargs):
        return [
            "theano_helpers.h",
            "ellip.h",
            "limbdark.h",
            "utils.h",
            "vector",
        ]

    def c_header_dirs(self, *args, **kwargs):
        dirs = [
            pkg_resources.resource_filename("starry", "_core/ops/lib/include")
        ]
        dirs += [
            pkg_resources.resource_filename(
                "starry", "_core/ops/lib/vendor/eigen_3.3.5"
            )
        ]
        return dirs

    def c_compile_args(self, *args, **kwargs):
        opts = ["-std=c++11", "-O2", "-DNDEBUG"]
        if sys.platform == "darwin":
            opts += ["-stdlib=libc++", "-mmacosx-version-min=10.7"]
        return opts

    def perform(self, *args):
        raise NotImplementedError("Only C op is implemented")



# -*- coding: utf-8 -*-
from ...compat import Apply, Op, tt
import numpy as np

__all__ = ["dotROp", "tensordotRzOp"]


class dotROp(Op):
    def __init__(self, func):
        self.func = func
        self._grad_op = dotRGradientOp(self)

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [tt.TensorType(inputs[0].dtype, (False, False))()]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return (shapes[0],)

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)

    def perform(self, node, inputs, outputs):
        outputs[0][0] = self.func(*inputs)

    def grad(self, inputs, gradients):
        return self._grad_op(*(inputs + gradients))


class dotRGradientOp(Op):
    def __init__(self, base_op):
        self.base_op = base_op

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [i.type() for i in inputs[:-1]]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return shapes[:-1]

    def perform(self, node, inputs, outputs):
        bM, bx, by, bz, btheta = self.base_op.func(*inputs)
        outputs[0][0] = np.reshape(bM, np.shape(inputs[0]))
        outputs[1][0] = np.reshape(bx, np.shape(inputs[1]))
        outputs[2][0] = np.reshape(by, np.shape(inputs[2]))
        outputs[3][0] = np.reshape(bz, np.shape(inputs[3]))
        outputs[4][0] = np.reshape(btheta, np.shape(inputs[4]))


class tensordotRzOp(Op):
    def __init__(self, func):
        self.func = func
        self._grad_op = tensordotRzGradientOp(self)

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [tt.TensorType(inputs[0].dtype, (False, False))()]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return [[shapes[1][0], shapes[0][-1]]]

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)

    def perform(self, node, inputs, outputs):
        outputs[0][0] = self.func(*inputs)

    def grad(self, inputs, gradients):
        return self._grad_op(*(inputs + gradients))


class tensordotRzGradientOp(Op):
    def __init__(self, base_op):
        self.base_op = base_op

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [i.type() for i in inputs[:-1]]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return shapes[:-1]

    def perform(self, node, inputs, outputs):
        bM, btheta = self.base_op.func(*inputs)
        outputs[0][0] = np.reshape(bM, np.shape(inputs[0]))
        outputs[1][0] = np.reshape(btheta, np.shape(inputs[1]))
